import numpy as np
from numpy import exp, pi
import scipy.constants as spc 


# Liquid water

def h2o_liquid_KleinSwift(frequency, temperature, salinity): 
    """ Model from Klein and Swift 1977 for sea water dielectric constant. 
        Valid for frequencies lower than 50 GHz.        

        :param frequency: Frequency in MHz
        :param temperature: Temperature in Kelvin
        :param salinity: Salinity in psu
        :return: Liquid water dielectric constant

    """ 

    SST = temperature - spc.zero_Celsius
    S = salinity
    f = frequency * 1e6

    e0 = 8.8541878128e-12  # spc.physical_constants['vacuum electric permittivity'][0]
    esw_inf = 4.9
    T = SST

    a = 1.0 + 1.613e-5 * T * S - 3.656e-3 * S + 3.210e-5 * S**2 - 4.232e-7 * S**3
    esw_T0 = 87.134 - 1.949e-1 * T - 1.276e-2 * T**2 + 2.491e-4 * T**3
    esw0 = esw_T0 * a

    tau_T0 = (1.1109e-10 - 3.824e-12 * T + 6.938e-14 * T**2 - 5.096e-16 * T**3) / (2 * pi)
    b = 1.0 + 2.282e-5 * T * S - 7.638e-4 * S - 7.760e-6 * S**2 + 1.105e-8 * S**3
    tau_sw = tau_T0 * b

    sigma_25 = S * (0.18252 - 1.4619e-3 * S + 2.093e-5 * S**2 - 1.282e-7 * S**3)
    delt = 25 - T
    phi_e = delt * (2.033e-2 + 1.266e-4 * delt + 2.464e-6 * delt**2 - S * (1.849e-5 - 2.551e-7 * delt + 2.551e-8 * delt**2))
    sigma = sigma_25 * exp(-phi_e)

    esw_real = esw_inf + (esw0 - esw_inf) / (1 + (2 * pi * f * tau_sw)**2)
    esw_img = (2 * pi * f * tau_sw * (esw0 - esw_inf)) / (1 + (2 * pi * f * tau_sw)**2) + sigma / (2 * pi * e0 * f)

    epsilon = esw_real - 1j * esw_img

    return epsilon


def h2o_liquid_MeissnerWentz(frequency, temperature, salinity): 
    """ Model from Meissner and Wentz 2004 & 2012 for sea water dielectric constant

        :param frequency: Frequency in MHz 
        :param temperature: Temperature in Kelvin 
        :param salinity: Salinity in psu 
        :return: Liquid water dielectric constant
    """

    s_ppth = salinity
    celsius = temperature - spc.zero_Celsius
    ghz = frequency / 1e3

    x = np.array([5.723, 2.2379e-2, -7.1237e-4, 5.0478, -7.0315e-2, 6.0059e-4, 3.6143, 2.8841e-2, 1.3652e-1, 1.4825e-3, 2.4166e-4])

    # Pure water parameters
    es = (3.70886e4 - 8.2168e1 * celsius) / (4.21854e2 + celsius)
    e1 = x[0] + x[1] * celsius + x[2] * celsius**2
    v1 = (45 + celsius) / (x[3] + x[4] * celsius + x[5] * celsius**2)
    einf = x[6] + x[7] * celsius
    v2 = (45 + celsius) / (x[8] + x[9] * celsius + x[10] * celsius**2)

    # Conductivity
    sigma_35 = 2.903602 + 8.607e-2 * celsius + 4.738817e-4 * celsius**2 - 2.991e-6 * celsius**3 + 4.3047e-9 * celsius**4
    r_15 = s_ppth * (37.5109 + 5.45216 * s_ppth + 1.4409e-2 * s_ppth**2) / (1004.75 + 182.283 * s_ppth + s_ppth**2)
    a0 = (6.9431 + 3.2841 * s_ppth - 9.9486e-2 * s_ppth**2) / (84.850 + 69.024 * s_ppth + s_ppth**2)
    a1 = 49.843 - 0.2276 * s_ppth + 0.198e-2 * s_ppth**2
    r_t_r_15 = 1 + (a0 * (celsius - 15)) / (a1 + celsius)
    sigma = sigma_35 * r_15 * r_t_r_15

    # Salt water parameters
    # Parameters from Meissner and Wentz 2012
    b = np.array([-3.33330e-3, 4.74868e-6, 0, 2.39357e-3, -3.13530e-5, 2.52477e-7, -6.28908e-3, 1.76032e-4, -9.22144e-5, -1.99723e-2, 1.81176e-4, -2.04265e-3, 1.57883e-4])
    d = np.array([0.23232e-2, -0.79208e-4, 0.36764e-5, -0.35594e-6, 0.89795e-8])  # Correct d3 coefficient, as there's a typo in Meissner and Wentz 2012
    v1 = v1 * (1 + s_ppth * (d[0] + d[1] * celsius + d[2] * celsius**2 + d[3] * celsius**3 + d[4] * celsius**4))
    # Parameters from Meissner and Wentz 2004
    # b = np.array([-3.56417e-3, 4.74868e-6, 1.15574e-5, 2.39357e-3, -3.13530e-5, 2.52477e-7, -6.28908e-3, 1.76032e-4, -9.22144e-5, -1.99723e-2, 1.81176e-4, -2.04265e-3, 1.57883e-4])
    # v1 = v1 * (1 + s_ppth * (b[3] + b[4] * celsius + b[5] * celsius**2))

    es = es * exp(b[0] * s_ppth + b[1] * s_ppth**2 + b[2] * celsius * s_ppth)   
    e1 = e1 * exp(b[6] * s_ppth + b[7] * s_ppth**2 + b[8] * celsius * s_ppth)
    v2 = v2 * (1 + s_ppth * (b[9] + b[10] * celsius))
    einf = einf * (1 + s_ppth * (b[11] + b[12] * celsius))
    v1_mat = ghz * 1 / v1
    v2_mat = ghz * 1 / v2
    epsilon = (es - e1) / (1 + 1j * v1_mat) + (e1 - einf) / (1 + 1j * v2_mat) + einf - 1j * sigma * 17.97510 / ghz

    return epsilon


def h2o_liquid_Ellison(frequency, temperature, salinity): 
    """ Model from Ellison as discussed by Maetzler 2006 - *Thermal Microwave Radiation: Applications for Remote Sensing*
        and Ulaby and Long 2014 - *Microwave Radar and Radiometric Remote Sensing* books.
        Similar formulation to the Meissner and Wentz model.
        Valid for 0-30 C, salinity less than 40 parts per thousand, frequencies lower than 1 THz.
        
        :param frequency: Frequency in MHz 
        :param temperature: Temperature in Kelvin 
        :param salinity: Salinity in psu 
        :return: Liquid water dielectric constant 
    """

    celsius = temperature - spc.zero_Celsius
    s_ppth = salinity
    ghz = frequency / 1e3 

    # Constants 
    # Spare zero added in the 'a' array so that my indexing matches that in Ulaby and Long
    a = np.array([0, 0.46606917e-2, -0.26087876e-4, -0.63926782e-5, 0.63000075e1, 0.26242021e-2, -0.42984155e-2, 0.34414691e-4, 0.17667420e-3, -0.20491560e-6, 0.58366888e3, 0.12684992e3, 0.69227972e-4, 0.38957681e-6, 0.30742330e3, 0.12634992e3, 0.37245044e1, 0.92609781e-2, -0.26093754e-1])
    ew_0 = 87.85306 * exp(-0.00456992 * celsius - a[1] * s_ppth - a[2] * s_ppth**2 - a[3] * s_ppth * celsius)
    ew_1 = a[4] * exp(-a[5] * celsius - a[6] * s_ppth - a[7] * s_ppth * celsius)
    tw_1 = (a[8] + a[9] * s_ppth) * exp(a[10] / (celsius + a[11]))
    tw_2 = (a[12] + a[13] * s_ppth) * exp(a[14] / (celsius + a[15]))
    ew_inf = a[16] + a[17] * celsius + a[18] * s_ppth

    # Conductivity 
    sigma_35 = 2.903602 + 8.607e-2 * celsius + 4.7338819e-4 * celsius**2 - 2.991e-6 * celsius**3 + 4.3041e-9 * celsius**4 
    ps = s_ppth * (37.5109 + 5.45216 * s_ppth + .014409 * s_ppth**2) / (1004.75 + 182.283 * s_ppth + s_ppth**2)  
    a0 = (6.9431 + 3.2841 * s_ppth - 0.099486 * s_ppth**2) / (84.85 + 69.024 * s_ppth + s_ppth**2)
    a1 = 49.843 - 0.2276 * s_ppth + 0.00198 * s_ppth**2
    qs = 1 + (a0 * (celsius - 15)) / (celsius + a1)
    sigma = sigma_35 * ps * qs

    # Double Debye
    eps_r = ew_inf + (ew_0 - ew_1) / (1 + (2 * pi * ghz * tw_1)**2) + (ew_1 - ew_inf) / (1 + (2 * pi * ghz * tw_2)**2)
    eps_i = 2 * pi * ghz * tw_1 * (ew_0 - ew_1) / (1 + (2 * pi * ghz * tw_1)**2) + 2 * pi * ghz * tw_2 * (ew_1 - ew_inf) / (1 + (2 * pi * ghz * tw_2)**2) + sigma * 17.97510 / ghz

    epsilon = eps_r - 1j * eps_i 

    return epsilon


def h2o_liquid_Boutin(frequency, temperature, salinity): 
    """ Model for seawater dielectric constant from Boutin et al. 2020. 
        
        :param frequency: Frequency in MHz 
        :param temperature: Temperature in Kelvin 
        :param salinity: Salinity in psu 
        :return: Liquid water dielectric constant 
    """

    celsius = temperature - spc.zero_Celsius
    s_ppth = salinity
    ghz = frequency / 1e3 
    f0 = 17.97510

    x = [5.7230e+00, 2.2379e-02, -7.1237e-04, 5.0478e+00, -7.0315e-02,
         6.0059e-04, 3.6143e+00, 2.8841e-02, 1.3652e-01, 1.4825e-03, 2.4166e-04]
    z = [-3.56417e-03, 4.74868e-06, 1.15574e-05, 2.39357e-03, -3.13530e-05, 
         2.52477e-07, -6.28908e-03, 1.76032e-04, -9.22144e-05, -1.99723e-02, 
         1.81176e-04, -2.04265e-03, 1.57883e-04]
    PP = [0.000000001749069, 0.000001088535951, -0.000038972693320, 0.003228077425434]

    if celsius < -30.16: celsius = -30.16  

    # Pure water
    e0 = (3.70886e4 - 8.2168e1 * celsius) / (4.21854e2 + celsius)
    e1 = x[0] + x[1] * celsius + x[2] * celsius**2
    n1 = (45.00 + celsius) / (x[3] + x[4] * celsius + x[5] * celsius**2)

    # saline water
    # conductivity [s/m] taken from stogryn et al.
    sig35 = 2.903602 + 8.60700e-2 * celsius + 4.738817e-4 * celsius**2 - 2.9910e-6 * celsius**3 + 4.3047e-9 * celsius**4
    r15 = s_ppth * (37.5109 + 5.45216 * s_ppth + 1.4409e-2 * s_ppth**2) / (1004.75 + 182.283 * s_ppth + s_ppth**2)
    alpha0 = (6.9431 + 3.2841 * s_ppth - 9.9486e-2 * s_ppth**2) / (84.850 + 69.024 * s_ppth + s_ppth**2)
    alpha1 = 49.843 - 0.2276 * s_ppth + 0.198e-2 * s_ppth**2
    rtr15 = 1.0 + (celsius - 15.0) * alpha0 / (alpha1 + celsius)

    sig = sig35 * r15 * rtr15

    fSST = PP[0] * celsius**3 + PP[1] * celsius**2 + PP[2] * celsius + PP[3]

    a0 = 1 - s_ppth * fSST
    e0s = a0 * e0    
    b1 = 1
    n1s = n1 * b1  
    # a1  = np.exp(z[6] * s + z[7] * s2 + z[8] * s * sst); 
    a1 = 1
    e1s = e1 * a1   
    epsr = (e0s - e1s) / (1.0 + 1j * (ghz / n1s)) + e1s - 1j * sig * f0 / ghz

    epsilon = np.real(epsr) - 1j * abs(np.imag(epsr))

    return epsilon


def h2o_liquid_Zhou(frequency, temperature, salinity): 
    """ Model for seawater dielectric constant from Zhou et al. 2021. 
        
        :param frequency: Frequency in MHz 
        :param temperature: Temperature in Kelvin 
        :param salinity: Salinity in psu 
        :return: Liquid water dielectric constant 
    """ 
    celsius = temperature - spc.zero_Celsius
    s_ppth = salinity
    ghz = frequency / 1e3 

    e_inf = 4.9 
    tau_f = np.array([1.75030e-11, -6.12993e-13, 1.24504e-14, -1.14927e-16])
    tau = tau_f[0] + tau_f[1] * celsius + tau_f[2] * celsius**2 + tau_f[3] * celsius**3
    esdw_f = np.array([8.80516e1, -4.01796e-1, -5.10271e-5, 2.55892e-5])
    esdw = esdw_f[0] + esdw_f[1] * celsius + esdw_f[2] * celsius**2 + esdw_f[3] * celsius**3
    rswdw_f = np.array([3.97185e-3, -2.49205e-5, -4.27558e-5, 3.92825e-7, 4.15350e-7])
    rswdw = 1 - s_ppth * (rswdw_f[0] + rswdw_f[1] * celsius + rswdw_f[2] * s_ppth + rswdw_f[3] * celsius * s_ppth + rswdw_f[4] * s_ppth**2)
    sig_f = np.array([9.50470e-2, -4.30858e-4, 2.16182e-6])
    sig = sig_f[0] * s_ppth + sig_f[1] * s_ppth**2 + sig_f[2] * s_ppth**3
    rsig_f = np.array([3.76017e-2, 6.32830e-5, 4.8342e-7, -3.97484e-4, 6.26522e-6])
    rsig = 1 + celsius * (rsig_f[0] + rsig_f[1] * celsius + rsig_f[2] * celsius**2 + rsig_f[3] * s_ppth + rsig_f[4] * s_ppth**2)
    true_sig = sig * rsig

    e0 = 8.8541878128e-12  # spc.physical_constants['vacuum electric permittivity'][0]
    # epsilon = e_inf + (esdw * rswdw - e_inf) / (1 + 1j * 2 * pi * ghz * tau) - 1j * true_sig / (2 * pi * ghz * e0)
    epsilon = e_inf + (esdw * rswdw - e_inf) / (1 + 1j * 2 * pi * 1e9 * ghz * tau) - 1j * true_sig / (2 * pi * ghz * 1e9 * e0)

    return epsilon


def foam_Stogryn(frequency, angle): 
    """ From the empirical polarized foam *emissivity* (not dielectric constant) model of Stogryn 1972

        :param frequency: Frequency in MHz
        :param angle: Satellite look angle in degrees off nadir

        Anguelova parameterization will be added in the future

    """
    angle[angle > 70] = 70
    ghz = frequency / 1e3
    emis_nadir = (208 + 1.29 * ghz) / 288 
    F_h = 1 - 1.748e-3 * angle - 7.336e-5 * angle**2 + 1.004e-7 * angle**3
    F_v = 1 - 9.946e-4 * angle + 3.218e-5 * angle**2 - 1.187e-6 * angle**3 + 7e-20 * angle**10
    emis_H = emis_nadir * F_h
    emis_V = emis_nadir * F_v

    return emis_H, emis_V


# Ice 

def h2o_ice_Maetzler(frequency, temperature, salinity):
    """ Computes permittivity of impure ice from Maetzler 2006. The model was developed for salinity around 0.013 PSU,
        and extrapolation is based on linear assumption to salinity, so over-extrapolation isn't recommended.

        :param frequency: Frequency in MHz 
        :param temperature: Temperature in Kelvin 
        :param salinity: Salinity in psu 
        :return: Ice dielectric constant 

        Code and documentation modified from the Snow Microwave Radiative Transfer Model (SMRT) 
        https://www.smrt-model.science/
    """

    # Modify imaginary component calculated for pure ice
    ghz = frequency / 1e3
    ereal = 3.1884 + 9.1e-4 * (temperature - spc.zero_Celsius)
    theta = 300.0 / temperature - 1.0
    alpha = (0.00504 + 0.0062 * theta) * exp(-22.1 * theta)
    b1 = 0.0207
    b2 = 1.16e-11
    b = 335.0
    deltabeta = exp(- 9.963 + 0.0372 * (temperature - spc.zero_Celsius))
    betam = (b1 / temperature) * (exp(b / temperature) / ((exp(b / temperature) - 1)**2)) + b2 * ghz**2
    beta = betam + deltabeta
    eimag = alpha / ghz + beta * ghz

    pure_ice_permittivity = ereal - 1j * eimag

    # Equation 5.37 from Maetzler 2006: Thermal Microwave Radiation: Applications for Remote Sensing
    g0 = 1866 * np.exp(-0.317 * ghz)
    g1 = 72.2 + 6.02 * ghz

    # Equation 5.36
    delta_Eimag = 1. / (g0 + g1 * (spc.zero_Celsius - temperature))

    # Equation 5.38
    S0 = 0.013 
    epsilon = pure_ice_permittivity - 1j * delta_Eimag * salinity / S0

    return epsilon 


# Soil

def soil_dobson(frequency, moisture, sand, clay, temperature=296, density=1.7): 
    """ Soil model from Dobson et al. 1985 as presented in Ulaby and Long 2014 

        :param frequency: Frequency in MHz
        :param moisture: Volumetric soil moisture in g/cm3
        :param sand: Sand fraction of soil
        :param clay: Clay fraction of soil
        :param temperature: Soil temperature in K
        :param density: Soil density in g/cm3
        :return: Soil dielectric constant

    """
    ghz = frequency / 1e3
    moist = moisture
    san = sand
    cla = clay

    # Constants 
    alpha = 0.65
    beta_1 = 1.27 - 0.519 * san - 0.152 * cla 
    beta_2 = 2.06 - 0.928 * san - 0.255 * cla 
    sigma = -1.65 + 1.939 * density - 2.256 * san + 1.594 * cla 
    eps_w = h2o_liquid_KleinSwift(frequency, temperature, 0)  # This can be replaced with a different water model from above if necessary.
    eps_r = np.real(eps_w)
    eps_i = -np.imag(eps_w) + np.nan_to_num((2.65 - density) / (2.65 * np.array([moist])), nan=0, posinf=0, neginf=0) * sigma * 17.97510 / ghz
    es_r = (1 + 0.66 * density + moist**beta_1 * eps_r**alpha - moist)**(1 / alpha)
    es_i = moist**beta_2 * eps_i

    epsilon = es_r - 1j * es_i

    return epsilon


def soil_peplinski(frequency, moisture, sand, clay, temperature=296, density=1.7): 
    """ Soil model from Dobson et al. 1985 with Peplinski 1995 conductivity term
        The modified conductivity term is specifically relevant for frequencies between 0.3 - 1.3 GHz 

        :param frequency: Frequency in MHz
        :param moisture: Volumetric soil moisture in g/cm3
        :param sand: Sand fraction of soil
        :param clay: Clay fraction of soil
        :param temperature: Soil temperature in K
        :param density: Soil density in g/cm3
        :return: Soil dielectric constant

    """
    ghz = frequency / 1e3
    moist = moisture
    san = sand
    cla = clay

    # Constants 
    alpha = 0.65
    beta_1 = 1.27 - 0.519 * san - 0.152 * cla 
    beta_2 = 2.06 - 0.928 * san - 0.255 * cla 
    sigma = 0.0467 + 0.22 * density - 0.411 * san + 0.661 * cla 
    eps_w = h2o_liquid_KleinSwift(frequency, temperature, 0) 
    eps_r = np.real(eps_w)
    eps_i = -np.imag(eps_w) + np.nan_to_num((2.65 - density) / (2.65 * np.array([moist])), nan=0, posinf=0, neginf=0) * sigma * 17.97510 / ghz

    es_r = (1 + 0.66 * density + moist**beta_1 * eps_r**alpha - moist)**(1 / alpha)
    es_i = moist**beta_2 * eps_i

    epsilon = es_r - 1j * es_i

    return epsilon


def soil_modified_dobson(frequency, moisture, sand, clay): 
    """ Soil model from Dobson et al. 1985 with modification from the Aquarius Algorithm
        Theoretical Basis Document. Performs better for wet soil 

        :param frequency: Frequency in MHz
        :param moisture: Volumetric soil moisture in g/cm3
        :param sand: Sand fraction of soil
        :param clay: Clay fraction of soil
        :return: Soil dielectric constant
    """
    ghz = frequency / 1e3
    beta = 1.09 - 0.11 * sand + 0.18 * clay
    alpha = 0.65
    r_b = 1.3  # Soil bulk density 
    r_ss = 2.65  # Density of solid soil
    eps_ss = 4.7  # Dielectric constant of dry soil
    fo = 18.64  # Relaxation frequency of GHz of pure water

    eps = 1 + r_b / r_ss * (eps_ss**alpha - 1) + moisture**beta * (4.9 + 74.1 / (1 - 6j * ghz / fo)**alpha) - moisture
    epsilon = np.real(eps**(1 / alpha)) - 1j * np.imag(eps**(1 / alpha))

    return epsilon


def soil_hallikainen(moisture, sand, clay): 
    """ L Band soil model from Hallikainen et al. 1985 as discussed in the Aquarius ATBD. 
        Performs better for dry soil. 

        :param moisture: Volumetric soil moisture in g/cm3
        :param sand: Sand fraction of soil
        :param clay: Clay fraction of soil
        :return: Soil dielectric constant
    """

    eps_r = 2.862 - 0.012 * sand + 0.001 * clay + moisture * (3.803 + 0.462 * sand - 0.341 * clay) + moisture**2 * (119.006 - 0.5 * sand + 0.663 * clay)
    eps_i = 0.356 - 0.003 * sand + 0.008 * clay + moisture * (5.507 + 0.044 * sand - 0.002 * clay) + moisture**2 * (17.753 - 0.313 * sand + 0.206 * clay)

    epsilon = eps_r - 1j * eps_i 

    return epsilon 


def soil_wang(moisture): 
    """ L Band soil model from Le Vine and Kao 1997 and Wang 1980. Performs better for wet soil 
        
        :param moisture: Volumetric soil moisture in g/cm3
        :return: Soil dielectric constant
    """

    eps_r = 3.1 + 17.36 * moisture + 63.12 * moisture**2
    eps_i = 0.031 + 4.65 * moisture + 20.42 * moisture**2

    epsilon = eps_r - 1j * eps_i 

    return epsilon



import numpy as np 
import scipy.constants as spc 
import scipy.special as sps 
import scipy.integrate as sint 


def read_antenna_pattern(file): 
    """ Reads a simple antenna or feed pattern (i.e. one pattern, no polarization dependence)
        from a text file and returns in a format for input to a FOAM spacecraft. 

        The file format is a csv style with dimensions of Nx3
        Column 1: Elevation/Theta (slowest varying index)
        Column 2: Azimuth/Phi (fasted varying index)
        Column 3: Normalized linear power pattern (Integral = 1)
    """ 

    ap = np.loadtxt(file, delimiter=',', skiprows=1)
    theta = ap[:, 0]
    phi = ap[:, 1]
    pattern = ap[:, 2]
    theta_grid = np.unique(theta)
    phi_grid = np.unique(phi)
    pattern_grid = pattern.reshape(len(theta_grid), len(phi_grid))

    ap_dict = {'theta': theta_grid, 'phi': phi_grid, 'pattern': pattern_grid}
    return ap_dict


def make_uniform_reflector(diameter, frequency, file):  
    """ Generates an antenna pattern assuming uniform illumination 
        of a circular reflector with an arbitrary diameter 

        :param diameter: Reflector diameter in meters
        :param frequency: Frequency in MHz 
        :param file: Output text file to write pattern to
    """ 

    wave = spc.c / (frequency * 1e6)  # In meters
    k = 2 * np.pi / wave
    theta_grid = np.radians(np.arange(0, 90, 0.2))
    azi_grid = np.radians(np.arange(-180, 180, 0.2))
    pattern_grid = np.zeros((len(theta_grid), len(azi_grid)))
    for i, thet in enumerate(theta_grid): 
        out = sint.quad(lambda r, th: sps.j0(k * r * np.sin(th)) * r, 0, diameter, args=(thet))
        pattern_grid[i, :] = 2 * np.pi * out[0]  # This is the E-field pattern

    dphi = np.mean(np.gradient(azi_grid))
    dth = np.mean(np.gradient(theta_grid))
    theta_pattern = np.sum(pattern_grid**2, axis=1)
    total_power = np.trapz(theta_pattern * np.sin(theta_grid) * dth * dphi)
    # Power pattern is sufficient for phase-incoherent signals 
    power_pattern = pattern_grid**2 / total_power

    am, tm = np.meshgrid(azi_grid, theta_grid)
    save_pattern = np.vstack([np.degrees(tm).flatten(), np.degrees(am).flatten(), power_pattern.flatten()]).T 
    np.savetxt(file, save_pattern, header='theta, phi, linear pattern', delimiter=',')





        



# geomag.py
# by Christopher Weiss cmweiss@gmail.com
# modified for numpy array support by Alex Akins for use in FOAM

# Adapted from the geomagc software and World Magnetic Model of the NOAA
# Satellite and Information Service, National Geophysical Data Center
# http://www.ngdc.noaa.gov/geomag/WMM/DoDWMM.shtml
#
# Suggestions for improvements are appreciated.

# USAGE:
#
# >>> gm = geomag.GeoMag("WMM.COF")
# >>> mag = gm.GeoMag(80,0)
# >>> mag.dec
# -6.1335150785195536
# >>>

import numpy as np
import os
from datetime import date


class GeoMag:

    def GeoMag(self, dlat, dlon, h=0, time=date.today()):  # latitude (decimal degrees), longitude (decimal degrees), altitude (feet), date

        # time = date('Y') + date('z')/365
        time = time.year + ((time - date(time.year, 1, 1)).days / 365.0)
        alt = h / 3280.8399

        shape = np.shape(dlat)
        if type(dlat) is not np.ndarray: 
            dlat = np.array(dlat).flatten()
        if type(dlon) is not np.ndarray: 
            dlon = np.array(dlon).flatten()
        if type(h) is not np.ndarray: 
            h = np.array(h).flatten()
        if type(time) is not np.ndarray: 
            time = np.array(time).flatten()

        # Allowing arrays to be stored as elements
        self.p = self.p.astype(object)
        self.dp = self.dp.astype(object)
        self.tc = self.tc.astype(object)
        self.pp = self.pp.astype(object)
        self.sp = self.sp.astype(object)
        self.cp = self.cp.astype(object)

        otime = oalt = olat = olon = -1000.0 * np.ones(np.shape(dlat))

        dt = time - self.epoch
        glat = dlat
        glon = dlon
        rlat = np.radians(glat)
        rlon = np.radians(glon)
        srlon = np.sin(rlon)
        srlat = np.sin(rlat)
        crlon = np.cos(rlon)
        crlat = np.cos(rlat)
        srlat2 = srlat * srlat
        crlat2 = crlat * crlat
        self.sp[1] = srlon
        self.cp[1] = crlon

        # /* CONVERT FROM GEODETIC COORDS. TO SPHERICAL COORDS. */
        if ((alt != oalt).any() or (glat != olat).any()):
            q = np.sqrt(self.a2 - self.c2 * srlat2)
            q1 = alt * q
            q2 = ((q1 + self.a2) / (q1 + self.b2)) * ((q1 + self.a2) / (q1 + self.b2))
            ct = srlat / np.sqrt(q2 * crlat2 + srlat2)
            st = np.sqrt(1.0 - (ct * ct))
            r2 = (alt * alt) + 2.0 * q1 + (self.a4 - self.c4 * srlat2) / (q * q)
            r = np.sqrt(r2)
            d = np.sqrt(self.a2 * crlat2 + self.b2 * srlat2)
            ca = (alt + d) / r
            sa = self.c2 * crlat * srlat / (r * d)

        if (glon != olon).any():
            for m in range(2, self.maxord + 1):
                self.sp[m] = self.sp[1] * self.cp[m - 1] + self.cp[1] * self.sp[m - 1]
                self.cp[m] = self.cp[1] * self.cp[m - 1] - self.sp[1] * self.sp[m - 1]

        aor = self.re / r
        ar = aor * aor
        br = bt = bp = bpp = 0.0
        for n in range(1, self.maxord + 1):
            ar = ar * aor

            # for (m=0,D3=1,D4=(n+m+D3)/D3;D4>0;D4--,m+=D3):
            m = 0
            D3 = 1
            # D4=(n+m+D3)/D3
            D4 = (n + m + 1)
            while D4 > 0:

                # /*
                # COMPUTE UNNORMALIZED ASSOCIATED LEGENDRE POLYNOMIALS
                # AND DERIVATIVES VIA RECURSION RELATIONS
                # */
                if ((alt != oalt).any() or (glat != olat).any()):
                    if (n == m):
                        self.p[m][n] = st * self.p[m - 1][n - 1]
                        self.dp[m][n] = st * self.dp[m - 1][n - 1] + ct * self.p[m - 1][n - 1]

                    elif (n == 1 and m == 0):
                        self.p[m][n] = ct * self.p[m][n - 1]
                        self.dp[m][n] = ct * self.dp[m][n - 1] - st * self.p[m][n - 1]

                    elif (n > 1 and n != m):
                        if (m > n - 2):
                            self.p[m][n - 2] = 0
                        if (m > n - 2):
                            self.dp[m][n - 2] = 0.0
                        self.p[m][n] = ct * self.p[m][n - 1] - self.k[m][n] * self.p[m][n - 2]
                        self.dp[m][n] = ct * self.dp[m][n - 1] - st * self.p[m][n - 1] - self.k[m][n] * self.dp[m][n - 2]

        # /*
                # TIME ADJUST THE GAUSS COEFFICIENTS
        # */
                if (time != otime).any():
                    self.tc[m][n] = self.c[m][n] + dt * self.cd[m][n]
                    if (m != 0):
                        self.tc[n][m - 1] = self.c[n][m - 1] + dt * self.cd[n][m - 1]

        # /*
                # ACCUMULATE TERMS OF THE SPHERICAL HARMONIC EXPANSIONS
        # */
                par = ar * self.p[m][n]

                if (m == 0):
                    temp1 = self.tc[m][n] * self.cp[m]
                    temp2 = self.tc[m][n] * self.sp[m]
                else:
                    temp1 = self.tc[m][n] * self.cp[m] + self.tc[n][m - 1] * self.sp[m]
                    temp2 = self.tc[m][n] * self.sp[m] - self.tc[n][m - 1] * self.cp[m]

                bt = bt - ar * temp1 * self.dp[m][n]
                bp = bp + (self.fm[m] * temp2 * par)
                br = br + (self.fn[n] * temp1 * par)
        # /*
                # SPECIAL CASE:  NORTH/SOUTH GEOGRAPHIC POLES
        # */
                if (st.any() == 0.0 and m == 1):
                    if (n == 1):
                        self.pp[n] = self.pp[n - 1]
                    else:
                        self.pp[n] = self.pp[n] + (ct * [st == 0.0] * 1 * self.pp[n - 1] - self.k[m][n] * self.pp[n - 2])
                    parp = ar * self.pp[n]
                    bpp = bpp + (self.fm[m] * temp2 * parp)

                D4 = D4 - 1
                m = m + 1

        if (st.any() == 0.0):
            bp[st == 0] = bpp[st == 0]
        else:
            bp = bp / (st + 1e-320)  # Prevents div by zero
        # /*
            # ROTATE MAGNETIC VECTOR COMPONENTS FROM SPHERICAL TO
            # GEODETIC COORDINATES
        # */
        bx = -bt * ca - br * sa
        by = bp
        bz = bt * sa - br * ca
        # /*
        # COMPUTE DECLINATION (DEC), INCLINATION (DIP) AND
        # TOTAL INTENSITY (TI)
        # */
        bh = np.sqrt((bx * bx) + (by * by))
        ti = np.sqrt((bh * bh) + (bz * bz))
        dec = np.degrees(np.arctan2(by, bx))
        dip = np.degrees(np.arctan2(bz, bh))
        # /*
        # COMPUTE MAGNETIC GRID VARIATION IF THE CURRENT
        # GEODETIC POSITION IS IN THE ARCTIC OR ANTARCTIC
        # (I.E. GLAT > +55 DEGREES OR GLAT < -55 DEGREES)

        # OTHERWISE, SET MAGNETIC GRID VARIATION TO -999.0
        # */

        # What does this code block do??
        # gv = -999.0
        # if (np.fabs(glat) >= 55.):
        #     if (glat > 0.0 and glon >= 0.0):
        #         gv = dec-glon
        #     if (glat > 0.0 and glon < 0.0):
        #         gv = dec+np.fabs(glon);
        #     if (glat < 0.0 and glon >= 0.0):
        #         gv = dec+glon
        #     if (glat < 0.0 and glon < 0.0):
        #         gv = dec-np.fabs(glon)
        #     if (gv > +180.0):
        #         gv = gv - 360.0
        #     if (gv < -180.0):
        #         gv = gv + 360.0

        otime = time
        oalt = alt
        olat = glat
        olon = glon

        class RetObj:
            pass
        retobj = RetObj()
        retobj.dec = dec
        retobj.dip = dip
        retobj.ti = ti
        retobj.bh = bh
        retobj.bx = bx.reshape(shape)
        retobj.by = by.reshape(shape)
        retobj.bz = bz.reshape(shape)
        retobj.lat = dlat
        retobj.lon = dlon
        retobj.alt = h
        retobj.time = time

        return retobj

    def __init__(self, wmm_filename=None):
        if not wmm_filename:
            wmm_filename = os.path.join(os.path.dirname(__file__), 'WMM.COF')
        wmm = []
        with open(wmm_filename) as wmm_file:
            for line in wmm_file:
                linevals = line.strip().split()
                if len(linevals) == 3:
                    self.epoch = float(linevals[0])
                    self.model = linevals[1]
                    self.modeldate = linevals[2]
                elif len(linevals) == 6:
                    linedict = {'n': int(float(linevals[0])),
                        'm': int(float(linevals[1])),
                        'gnm': float(linevals[2]),
                        'hnm': float(linevals[3]),
                        'dgnm': float(linevals[4]),
                        'dhnm': float(linevals[5])}
                    wmm.append(linedict)

        z = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.maxord = self.maxdeg = 12
        self.tc = np.array([z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13]])
        self.sp = np.array(z[0:14])
        self.cp = np.array(z[0:14])
        self.cp[0] = 1.0
        self.pp = np.array(z[0:13])
        self.pp[0] = 1.0
        self.p = np.array([z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14]])
        self.p[0][0] = 1.0
        self.dp = np.array([z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13]])
        self.a = 6378.137
        self.b = 6356.7523142
        self.re = 6371.2
        self.a2 = self.a * self.a
        self.b2 = self.b * self.b
        self.c2 = self.a2 - self.b2
        self.a4 = self.a2 * self.a2
        self.b4 = self.b2 * self.b2
        self.c4 = self.a4 - self.b4

        self.c = np.array([z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14]])
        self.cd = np.array([z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14], z[0:14]])

        for wmmnm in wmm:
            m = wmmnm['m']
            n = wmmnm['n']
            gnm = wmmnm['gnm']
            hnm = wmmnm['hnm']
            dgnm = wmmnm['dgnm']
            dhnm = wmmnm['dhnm']
            if (m <= n):
                self.c[m][n] = gnm
                self.cd[m][n] = dgnm
                if (m != 0):
                    self.c[n][m - 1] = hnm
                    self.cd[n][m - 1] = dhnm

        # /* CONVERT SCHMIDT NORMALIZED GAUSS COEFFICIENTS TO UNNORMALIZED */
        self.snorm = np.array([z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13]])
        self.snorm[0][0] = 1.0
        self.k = np.array([z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13], z[0:13]])
        self.k[1][1] = 0.0
        self.fn = np.array([0.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0])
        self.fm = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0])
        for n in range(1, self.maxord + 1):
            self.snorm[0][n] = self.snorm[0][n - 1] * (2.0 * n - 1) / n
            j = 2.0
            #for (m=0,D1=1,D2=(n-m+D1)/D1;D2>0;D2--,m+=D1):
            m = 0
            D1 = 1
            D2 = (n - m + D1) / D1
            while (D2 > 0):
                self.k[m][n] = (((n - 1) * (n - 1)) - (m * m)) / ((2.0 * n - 1) * (2.0 * n - 3.0))
                if (m > 0):
                    flnmj = ((n - m + 1.0) * j) / (n + m)
                    self.snorm[m][n] = self.snorm[m - 1][n] * np.sqrt(flnmj)
                    j = 1.0
                    self.c[n][m - 1] = self.snorm[m][n] * self.c[n][m - 1]
                    self.cd[n][m - 1] = self.snorm[m][n] * self.cd[n][m - 1]
                self.c[m][n] = self.snorm[m][n] * self.c[m][n]
                self.cd[m][n] = self.snorm[m][n] * self.cd[m][n]
                D2 = D2 - 1
                m = m + D1



import os
import re
from itertools import repeat
from datetime import datetime
from multiprocessing import Pool
import copy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt 
import cartopy.crs as ccrs 
import spiceypy as spice 
from tqdm import tqdm

from .utils.mk import manual_furnish
from .utils.config import dir_path, cache_path
spice_path = os.path.join(cache_path, 'spice')


class spacecraft(): 
    """ Class defining spacecraft instrument and pointing characteristics for 
        for ocean remote sensing on Earth. Each spacecraft object is numbered 
        as a member of a constellation, and each spacecraft can be initialized 
        with several microwave instruments.  

        :param sc_number: Integer identifier of the spacecraft object within a constellation 

        The NAIF ID of a spacecraft in SPICE is -1500 - sc_number
    """

    # Geophysical constants are defined in the geophys.ker file located in assets/ and discussed in SPICE docs (e.g. evlin)
    # The spkw10 documentation is wrong about the order of the constants, and it should follow the 
    # convention in geophys.ker and below
    geophys = np.array([1.082616e-3, -2.53881e-6, -1.65597e-6, 7.43669161e-2, 120, 78, 6378.135, 1])
    mu = 3.986004418e14
    kernel_epoch_range = [-3155716758.8160305, 1577880069.1839132]
    earth_mean_radius = 6371.0088e3

    def __init__(self, sc_number=0):
        self.sc_number = sc_number  # Spacecraft number, relevant if building a constellation
        self.naif_id = -1500 - sc_number
        self.rad_number = 0  # Running count of spacecraft radiometers
    
    def close(self): 
        self.unload()
        
    def furnish(self): 
        manual_furnish() 
        spice.furnsh(self.spk_path)
        spice.furnsh(self.sclk_path)
        spice.furnsh(self.fk_path)
        spice.furnsh(self.ck_path)
        for rad_path in (self.radiometer_fk_paths + self.radiometer_ck_paths): 
            spice.furnsh(rad_path)

    def unload(self): 
        spice.clpool() 

    def write_tle_kernels(self, file=None, elems=None, tle_epoch=None, start_epoch=0., end_epoch=100., epoch_res=10):
        """ Writing kernels for an orbiting spacecraft. Written kernels include spk, sclk, ck, fk.
            Separate CKs must be written for each radiometer

            :param file: File name of the TLE file used to initialized kernels. Must be specified if mode='file'
            :param elems: List of orbital elements (see spacecraft.get_manual_elems)
            :param tle_epoch: Epoch for elements (see spacecraft.get_manual_elems)
            :param start_epoch: Starting epoch of the spacecraft operation period in seconds past J2000
            :param end_epoch: Ending epoch of the spacecraft operation period in seconds past J2000
            :param epoch_res: Time resolution for kernel calculations in seconds

            Spacecraft CK coordinate system is X -> toward planet, Y -> Orbital velocity vector
        """

        manual_furnish()
        # Start and end epochs are extended to the outermost second  
        start_epoch = np.floor(start_epoch)
        end_epoch = np.ceil(end_epoch)

        self.start_epoch = start_epoch 
        self.end_epoch = end_epoch
        self.epoch_res = epoch_res

        if file is not None: 
            tle_epoch, elems = self.read_tle(file)
        elif (elems is not None) and (tle_epoch is not None): 
            pass 
        else: 
            raise RuntimeError('Either a filename or an elements list must be provided')
        self.orbital_period = 2 * np.pi / (elems[-2] / 60)
        # Writing to SPK
        self.spk_path = os.path.join(spice_path, 'spacecraft_{}.bsp'.format(self.sc_number))
        if os.path.exists(self.spk_path): os.remove(self.spk_path)
        file_spk = spice.spkopn(self.spk_path, 'SPACECRAFT_{}'.format(self.sc_number), 0)
        # Reference frame is important, and docs say it should be J2000 (for mkspk)
        spice.spkw10(file_spk, self.naif_id, 399, 'J2000', 
                     self.kernel_epoch_range[0], self.kernel_epoch_range[1], 
                     'SPACECRAFT_{}'.format(self.sc_number), self.geophys, 1, elems, [tle_epoch])
        spice.spkcls(file_spk)
        spice.furnsh(self.spk_path)

        # Writing the SCLK
        # Note: For generality, the SCLK string format is seconds:microseconds
        # This constrains the integration times to microsecond resolution

        now = datetime.now().strftime('@%Y-%m-%d/%H:%M:%S')  # Unique indicator of SCLK is time of definition
        self.sclk_path = os.path.join(spice_path, 'spacecraft_{}.sclk'.format(self.sc_number))
        if os.path.exists(self.sclk_path): os.remove(self.sclk_path)
        file_sclk = open(self.sclk_path, 'w')
        # SCLK Kernel ID: Unique idenfitier, set as the current date 
        # SCLK Data Type: Always type 1
        # SCLK Time System: 1 is BDT and 2 is TDT (see sclk.req)
        # SCLK N Fields: Number of clock fields (e.g. major and minor ticks would be two fields)
        # SCLK Moduli: Modulus value of the fields, or after which value they restart
        # SCLK Offsets: Offsets of clock on initialization 
        # SCLK Output Delim: 1 is '.', 2 is ':', 3 is '-', 4 is ',', and 5 is ' '
        # SCLK Partition: SCLK tick values at the start and end of a time partition. FOAM spacecrafts
        #                 will only have one partition, so this corresponds to the final tick
        # SCLK Coefficients: Triplets of (SCLK values (ticks), parallel time values, rates (same unit as time values))

        diff = end_epoch - start_epoch
        file_sclk.write("""KPL/SCLK
                    \\begindata
                    SCLK_KERNEL_ID              = {1}
                    SCLK_DATA_TYPE_{0}          = 1
                    SCLK01_TIME_SYSTEM_{0}      = 1
                    SCLK01_N_FIELDS_{0}         = 2
                    SCLK01_MODULI_{0}           = ({2}, {3})
                    SCLK01_OFFSETS_{0}          = (0, 0)
                    SCLK01_OUTPUT_DELIM_{0}     = 2

                    SCLK_PARTITION_START_{0}    = 0
                    SCLK_PARTITION_END_{0}      = {4}
                    SCLK01_COEFFICIENTS_{0}     = (

                    0.0000000000000E+00     {5:.13E}     1.0000000000000E+00                           
                    {4:.13E}     {6:.13E}     1.0000000000000E+00 )                            
                    """.format(abs(self.naif_id),
                               now, 
                               int(diff),  # seconds 
                               int(1e6),  # microseconds
                               int(diff * 1e6),  # Clock ticks
                               int(start_epoch),  # seconds
                               int(end_epoch)))  # seconds
                               
        file_sclk.close()
        spice.furnsh(self.sclk_path)

        # Writing the FK
        self.fk_path = os.path.join(spice_path, 'spacecraft_{}.tf'.format(self.sc_number))
        if os.path.exists(self.fk_path): os.remove(self.fk_path)
        file_fk = open(self.fk_path, 'w')
        # FK Frame ID: Same as the CLASS ID for a CK 
        # FK Name: Name of the frame 
        # FK Class: CK frame is class 3 
        # FK Class ID: Integer used internally by SPICE to address the frame 
        #              Safe range specified by NAIF is 1400000 - 2000000
        # FK Center: Center of the frame 
        # SCLK and SPK are specified via reference to respective naif IDs (which are the same for FOAM)
        file_fk.write("""KPL/FK 
                    \\begindata
                    FRAME_SPACECRAFT_{2}        = {0}
                    FRAME_{0}_NAME              = 'SPACECRAFT_{2}'
                    FRAME_{0}_CLASS             = 3
                    FRAME_{0}_CLASS_ID          = {0}
                    FRAME_{0}_CENTER            = {1}
                    CK_{0}_SCLK                 = {1}
                    CK_{0}_SPK                  = {1}
                    """.format(int(self.naif_id * 1e3), self.naif_id, self.sc_number))    
        file_fk.close()
        spice.furnsh(self.fk_path)

        # Writing the spacecraft CK
        # Coordinate system is X -> toward planet, Y -> Orbital velocity vector
        self.ck_path = os.path.join(spice_path, 'spacecraft_{}.bck'.format(self.sc_number))
        if os.path.exists(self.ck_path): os.remove(self.ck_path)
        enc_start = spice.sce2t(self.naif_id, start_epoch)
        enc_stop = spice.sce2t(self.naif_id, end_epoch)

        # Procedures for Type 3 specification
        sepochs = np.arange(start_epoch, end_epoch, epoch_res) 
        sepochs[-1] = end_epoch
        qts = []
        sncs = []
        for i in range(len(sepochs)): 
            starg, light = spice.spkezr('{}'.format(self.naif_id), sepochs[i], 'J2000', 'NONE', '399')
            npos = -starg[:3]  # X vector
            velvec = starg[3:]  # Y vector
            cmat = spice.twovec(spice.vhat(npos), 1, spice.vhat(velvec), 2)

            # If initializing from scratch this is ok, but not so if using an actual TLE
            # lon, lat, tall = spice.recgeo(pos, 6378.137, 1 / 298.257223563)  # WGS84 ellipsoid
            # rm_lat = spice.rotate(-lat, 2)
            # rm_lon = spice.rotate(np.pi - lon, 3)
            # cmat = rm_lon @ rm_lat 
            # qts.append(spice.m2q(cmat.T.copy()))

            qts.append(spice.m2q(cmat.copy()))
            sncs.append(spice.sce2t(self.naif_id, sepochs[i]))

        qts = np.squeeze(np.array([qts]))
        sncs = np.squeeze(np.array([sncs]))
        avvs = np.zeros((len(sncs), 3))
        file_ck = spice.ckopn(self.ck_path, 'SPACECRAFT_{}'.format(self.sc_number), 0)
        spice.ckw03(file_ck, enc_start, enc_stop, int(self.naif_id * 1e3), 'J2000', True, 'SPACECRAFT_{}'.format(self.sc_number),
                    len(sncs), sncs, qts, avvs, 1, np.array([enc_start]))
        spice.ckcls(file_ck)
        spice.furnsh(self.ck_path)

        # Procedures for Type 2 specification
        # Generally aren't good past a day, but I'm leaving them here for now...
        #
        # inclination = elems[3]
        # raan = elems[4]
        # meanmotion = elems[8] / 60
        # pos, light = spice.spkpos('-999', start_epoch, 'J2000', 'LT+S', '399')
        # lon, lat, tall = spice.recgeo(pos, 6378.137, 1 / 298.257223563)  # WGS84 ellipsoid
        # # radius, lon, lat = spice.reclat(pos)
        # rm_lat = spice.rotate(-lat, 2)
        # rm_lon = spice.rotate(np.pi - lon, 3)
        # cmat = rm_lon @ rm_lat 
        # start_qt = spice.m2q(cmat.T.copy())
        # raan_vec = spice.rotvec(np.array([1, 0, 0]), -raan, 3)
        # vel_vec = spice.vrotv(np.array([0, 0, 1]), raan_vec, inclination)
        # ang_vel_vec = meanmotion * vel_vec  # In base reference frame
        # # Check if the vel vec and cmat vec are normal to each other 
        # # c_vec = cmat @ np.array([-1, 0, 0])
        # # print('Normal check: {0}'.format(np.dot(c_vec, vel_vec)))
        # file_ck = spice.ckopn(dir_path + 'assets/spice/current_spacecraft.bck', 'SPACECRAFT', 0)
        # spice.ckw02(file_ck, enc_start, enc_stop, -999000, 'J2000', 'SPACECRAFT', 1, np.array([enc_start]), np.array([enc_stop]),
        #             start_qt, ang_vel_vec, [int_time * 1e-3])
        # spice.ckcls(file_ck)
        # spice.furnsh(spice_path + 'current_spacecraft.bck')

    def write_radiometer_ck(self, look_angle, look_axis, rpm, scan_axis, antenna_pattern=None):
        """ Writes the camera kernel for a given radiometer. 
            This can be called multiple times with different ck numbers 
            to create several radiometers
            
            :param look_angle: Look angle in degrees 
            :param look_axis: Reference axis for look angle rotation
                              Either 'X', 'Y', 'Z' in spacecraft frame or a vector
                              A negative sign can also be added e.g. -X
            :param rpm: Revolutions per minute of scan platform.
            :param scan_axis: Scan rotation axis
                              Either 'X', 'Y', 'Z' in spacecraft frame or a vector
                              A negative sign can also be added e.g. -X
            :param antenna_pattern: Text file describing radiometer antenna pattern
                                    This is a placeholder for future functionality.

            Spacecraft CK coordinate system is X -> toward planet, Y -> Orbital velocity vector
            and radiometer CK is defined relative to the spacecraft system using the input arguments
        """ 
        start_epoch = self.start_epoch
        end_epoch = self.end_epoch

        # Check list of radiometer_ck_paths
        if hasattr(self, 'radiometer_ck_paths'): 
            last = self.radiometer_ck_paths[-1]
            last = re.split('[_.]', last)
            count = int(last[-2]) 
            count += 1
        else:
            self.radiometer_fk_paths = [] 
            self.radiometer_ck_paths = []
            count = 0

        # Select radiometer number by searching kernel pool 
        # for existing radiometer kernels   
        # count = 0
        # stay_flag = True 
        # while stay_flag: 
        #     string = 'spacecraft_{0}_radiometer_{1}.bck'.format(self.sc_number, count)
        #     found_flag = False
        #     num = spice.ktotal('all')
        #     for i in range(0, num): 
        #         nm, tp, src, hnd = spice.kdata(i, 'all')
        #         if string in nm: 
        #             found_flag = True
        #     if found_flag: 
        #         count += 1 
        #     else: 
        #         stay_flag = False

        # Writing the FK
        fk_path = os.path.join(spice_path, 'spacecraft_{0}_radiometer_{1}.tf'.format(self.sc_number, count))
        if os.path.exists(fk_path): os.remove(fk_path)
        file_fk = open(fk_path, 'w')
        file_fk.write("""KPL/FK 
                    \\begindata
                    FRAME_SPACECRAFT_{3}_RADIOMETER_{1}    = {0}
                    FRAME_{0}_NAME          = 'SPACECRAFT_{3}_RADIOMETER_{1}'
                    FRAME_{0}_CLASS         = 3
                    FRAME_{0}_CLASS_ID      = {0}
                    FRAME_{0}_CENTER        = {2}
                    CK_{0}_SCLK             = {2}
                    CK_{0}_SPK              = {2}
                    """.format(int(self.naif_id * 1e3 - 1 - count), count, self.naif_id, self.sc_number))    
        file_fk.close()
        spice.furnsh(fk_path)
        self.radiometer_fk_paths.append(fk_path)

        ck_path = os.path.join(spice_path, 'spacecraft_{0}_radiometer_{1}.bck'.format(self.sc_number, count))
        if os.path.exists(ck_path): os.remove(ck_path)
        enc_start = spice.sce2t(self.naif_id, start_epoch)
        enc_stop = spice.sce2t(self.naif_id, end_epoch)
        file_ck = spice.ckopn(ck_path, 'SPACECRAFT_{0}_RADIOMETER_{1}'.format(self.sc_number, count), 0)
        look_angle = np.radians(look_angle)
        
        if str(look_axis) == 'X': 
            look_axis = np.array([1, 0, 0])
        elif str(look_axis) == '-X': 
            look_axis = np.array([-1, 0, 0])
        elif str(look_axis) == 'Y': 
            look_axis = np.array([0, 1, 0]) 
        elif str(look_axis) == '-Y': 
            look_axis = np.array([0, -1, 0]) 
        elif str(look_axis) == 'Z': 
            look_axis = np.array([0, 0, 1])
        elif str(look_axis) == '-Z': 
            look_axis = np.array([0, 0, -1])
        rm = spice.axisar(look_axis, look_angle)
        start_qt = spice.m2q(rm)

        if str(scan_axis) == 'X': 
            scan_axis = np.array([1, 0, 0])
        elif str(scan_axis) == '-X': 
            scan_axis = np.array([-1, 0, 0])
        elif str(scan_axis) == 'Y': 
            scan_axis = np.array([0, 1, 0]) 
        elif str(scan_axis) == '-Y': 
            scan_axis = np.array([0, -1, 0]) 
        elif str(scan_axis) == 'Z': 
            scan_axis = np.array([0, 0, 1])
        elif str(scan_axis) == '-Z': 
            scan_axis = np.array([0, 0, -1])
        scan_axis = scan_axis
        ang_vel = -rpm * 2 * np.pi / 60 
        ang_vel_vec = ang_vel * scan_axis
        spice.ckw02(file_ck, enc_start, enc_stop, int(self.naif_id * 1e3 - 1 - count), 
                    'SPACECRAFT_{}'.format(self.sc_number), 
                    'SPACECRAFT_{0}_RADIOMETER_{1}'.format(self.sc_number, count), 1,
                    np.array([enc_start]), np.array([enc_stop]), start_qt, ang_vel_vec, [1e-6])
        spice.ckcls(file_ck)
        spice.furnsh(ck_path)
        self.radiometer_ck_paths.append(ck_path)
        self.rad_number += 1

    def write_stationary_kernels(self, sub_long, sub_lat, height): 
        """ Writes SPICE kernels for a stationary spacecraft with a single pointing

            :param sub_long: Sub-Observer longitude in degrees 
            :param sub_lat: Sub-Observer latitude in degrees
            :param height: Height above sub-observer point in km

            This routine is currently broken, and will raise an error if called.  
        """

        raise RuntimeError('This kernel procedure is out of date')
        sub_lat = np.radians(sub_lat)
        sub_long = np.radians(sub_long)
        os.system('rm ' + dir_path + 'assets/spice/current_spacecraft.bsp')
        file = spice.spkopn(dir_path + 'assets/spice/current_spacecraft.bsp', 'SPACECRAFT', 0)
        rect = spice.latrec(self.earth_mean_radius / 1e3 + height, sub_long, sub_lat)
        spice.spkw09(file, -999, 399, 'ITRF93', 
            self.kernel_epoch_range[0], self.kernel_epoch_range[1],  # 1900 and 2500 epochs
            'Spacecraft Ephemeris', 1, 2,
            [[rect[0], rect[1], rect[2], 0, 0, 0], 
            [rect[0], rect[1], rect[2], 0, 0, 0]], 
            [self.kernel_epoch_range[0], self.kernel_epoch_range[1]])
        spice.spkcls(file)
        os.system('rm ' + dir_path + 'assets/spice/current_spacecraft.tf')
        file = open(dir_path + 'assets/spice/current_spacecraft.tf', 'w')
        # Write fixed offset reference frame
        # X towards planet, Y left, Z up 
        # Check ITRF image in spice docs
        file.write("""KPL/FK 
                    \\begindata
                    FRAME_SPACECRAFT            = -999000
                    FRAME_-999000_NAME          = 'SPACECRAFT'
                    FRAME_-999000_CLASS         = 4
                    FRAME_-999000_CLASS_ID      = -999000
                    FRAME_-999000_CENTER        = -999
                    OBJECT_-999_FRAME           = -999000
                    TKFRAME_-999000_SPEC     = 'ANGLES'
                    TKFRAME_-999000_RELATIVE = 'ITRF93'
                    TKFRAME_-999000_ANGLES   = ( {}, {}, {} )
                    TKFRAME_-999000_AXES     = (       3,        2,   3 )
                    TKFRAME_-999000_UNITS    = 'RADIANS'
                    """.format(np.pi, -sub_lat, -sub_long))
        file.close()
        spice.furnsh(dir_path + 'assets/spice/current_spacecraft.bsp')
        spice.furnsh(dir_path + 'assets/spice/current_spacecraft.tf')

    def make_track_grid(self, start_epoch, end_epoch, resolution):
        """ Generates a sequence of sub-spacecraft latitutdes and longitudes 
            
            :param start_epoch: Observation start epoch
            :param end_epoch: Observation end epoch
            :param resolution: Time resolution of tracks

            :return: A dictionary of longitudes, latitudes, and epochs 
        """
        epochs = np.arange(start_epoch, end_epoch, resolution) 
        lons = np.zeros(len(epochs))
        lats = np.zeros(len(epochs))
        for i in range(len(epochs)):
            try: 
                spoint, trgepc, srfvec = spice.sincpt('ELLIPSOID', 'EARTH', epochs[i], 'ITRF93', 'NONE', '{}'.format(self.naif_id), 
                                                      'SPACECRAFT_{}'.format(self.sc_number), np.array([1, 0, 0]))
            except spice.utils.exceptions.NotFoundError: 
                spoint = np.array([1, 0, 0])
                print('WARNING: No intercept determined for epoch {}, setting intercept to zero'.format(epochs[i]))
            lon, lat, tall = spice.recgeo(spoint, 6378.137, 1 / 298.257223563)  # WGS84 ellipsoid
            lons[i] = np.degrees(lon)
            lats[i] = np.degrees(lat)

        track_dict = {'epoch': epochs, 'lon': lons, 'lat': lats}

        return track_dict       

    def make_obs_grid(self, start_epoch, end_epoch, resolution, bounds=None):
        """ Generates a sequence of latitudes, longitudes and angles of observations between the
            start and end epochs with a given time resolution for each radiometer instrument 
            on the spacecraft 

            :param start_epoch: Observation start epoch
            :param end_epoch: Observation end epoch
            :param resolution: Time resolution of tracks
            :param bounds: Tuple of tuples ((lower_lat, upper_lat), (lower_lon, upper_lon))
                           Default is None

            :return: A dictionary of epochs, longitudes, latitudes, incidence elevations (theta), 
                    incidence azimuths (phi), right ascension and declination in the direction of 
                    specular reflection, and flags for if the sun and moon are in that direction 

                    The shape of the arrays in each dictionary value (except epochs) is NxM where N is the number 
                    of radiometers on the spacecraft, and M is the length of the epochs value

        """ 

        epochs = np.arange(start_epoch, end_epoch, resolution) 
        if bounds is None: 
            obs_dict = self._mog(epochs)
        else: 
            obs_dict = self._mog_bound(epochs, bounds)

        return obs_dict

    def _mog(self, epochs, ith=0): 
        """ Grid routine taking a single argument of epochs
        """

        # Commented blocks are a bit slower, but more legible 
        # lons = np.zeros((self.rad_number, len(epochs)))
        # lats = np.zeros((self.rad_number, len(epochs)))
        # thetas = np.zeros((self.rad_number, len(epochs)))
        # phis = np.zeros((self.rad_number, len(epochs)))
        # ras = np.zeros((self.rad_number, len(epochs)))
        # decs = np.zeros((self.rad_number, len(epochs)))
        # sun_flag = np.zeros((self.rad_number, len(epochs)), dtype=bool)
        # moon_flag = np.zeros((self.rad_number, len(epochs)), dtype=bool)

        splist = list(np.zeros(self.rad_number * len(epochs)))
        srlist = list(np.zeros(self.rad_number * len(epochs)))

        for i, eps in enumerate(tqdm(epochs, colour='blue', desc='Thread {}'.format(ith))):
            for j in range(self.rad_number):
                try: 
                    spoint, trgepc, srfvec = spice.sincpt('ELLIPSOID', 'EARTH', eps, 'ITRF93', 'NONE', '{}'.format(self.naif_id), 
                                                          'SPACECRAFT_{0}_RADIOMETER_{1}'.format(self.sc_number, j), np.array([1, 0, 0]))
                except spice.utils.exceptions.NotFoundError: 
                    spoint = np.array([1, 0, 0])
                    srfvec = np.array([1, 0, 0])
                    print('WARNING: No intercept determined for epoch {}, setting intercept to zero'.format(eps))
                
                # Run in loop
                # lon, lat, theta, phi, ra, dec, found_sun, found_moon = self.get_intercept_params(spoint, srfvec, eps)
                # lons[j, i] = np.degrees(lon)
                # lats[j, i] = np.degrees(lat)
                # thetas[j, i] = np.degrees(theta)
                # phis[j, i] = np.degrees(phi)
                # ras[j, i] = np.degrees(ra)
                # decs[j, i] = np.degrees(dec)
                # sun_flag[j, i] = found_sun 
                # moon_flag[j, i] = found_moon 

                splist[i * self.rad_number + j] = spoint
                srlist[i * self.rad_number + j] = srfvec

        # External list comps, harder to read but a bit faster
        zip_eps = epochs[:, np.newaxis] * np.ones(self.rad_number)  # Pad epochs
        zipper = list(zip(splist, srlist, zip_eps.flatten()))  # Zip them up
        pack = np.array([self.get_intercept_params(x[0], x[1], x[2]) for x in zipper], dtype=object)  # Get param pack
        keys = ['epoch', 'lon', 'lat', 'theta', 'phi', 'ra', 'dec', 'sun_flag', 'moon_flag']
        vals = [np.degrees(pack[:, x].astype(float).reshape(len(epochs), self.rad_number).T) for x in range(6)]  # Make obs dict
        vals2 = [pack[:, x].astype(bool).reshape(len(epochs), self.rad_number).T for x in [6, 7]]
        
        vals = [zip_eps.T] + vals + vals2

        obs_dict = dict(zip(keys, vals))

        # obs_dict = {'epoch': epochs, 'lon': lons, 'lat': lats, 'theta': thetas, 'phi': phis, 
        #             'ra': ras, 'dec': decs, 'sun_flag': sun_flag, 'moon_flag': moon_flag}

        return obs_dict

    def _mog_bound(self, epochs, bounds, ith=0): 
        """ Grid routine taking arguments of epochs and bounds 
            Separate since the act of latitude checking may 
            reduce performance in the bound-free case
        """

        splist = list(np.zeros(self.rad_number * len(epochs)))
        srlist = list(np.zeros(self.rad_number * len(epochs)))
        lat_bounds = bounds[0]
        lon_bounds = bounds[1]
        for i, eps in enumerate(tqdm(epochs, colour='blue', desc='Thread {}'.format(ith))):
            pos, lt = spice.spkpos(str(self.naif_id), eps, 'ITRF93', 'NONE', 'EARTH')
            lon, lat, tall = spice.recgeo(pos, 6378.137, 1 / 298.257223563)  # WGS84 ellipsoid
            lon = np.degrees(lon)
            lat = np.degrees(lat)
            if (lon > lon_bounds[0]) and (lon < lon_bounds[1]) and (lat > lat_bounds[0]) and (lat < lat_bounds[1]): 
                for j in range(self.rad_number):
                    try: 
                        spoint, trgepc, srfvec = spice.sincpt('ELLIPSOID', 'EARTH', eps, 'ITRF93', 'NONE', '{}'.format(self.naif_id), 
                                                              'SPACECRAFT_{0}_RADIOMETER_{1}'.format(self.sc_number, j), np.array([1, 0, 0]))
                    except spice.utils.exceptions.NotFoundError: 
                        spoint = np.array([1, 0, 0])
                        srfvec = np.array([1, 0, 0])
                        print('WARNING: No intercept determined for epoch {}, setting intercept to zero'.format(eps))

                    splist[i * self.rad_number + j] = spoint
                    srlist[i * self.rad_number + j] = srfvec

        # Thin zero valued lists
        rlsp = range(len(splist))
        mask = np.array([True if type(splist[x]) is np.ndarray else False for x in rlsp])
        splist = [splist[x] for x in rlsp if type(splist[x]) is np.ndarray]
        srlist = [srlist[x] for x in rlsp if type(srlist[x]) is np.ndarray]
        # Do all epochs check out? 
        mask = np.bitwise_and.reduce(mask.reshape(len(epochs), self.rad_number), axis=1)
        epochs = epochs[mask]
        if (mask == False).all():
            return {}
        else:  
            # External list comps, harder to read but a bit faster
            zip_eps = epochs[:, np.newaxis] * np.ones(self.rad_number)  # Pad epochs
            zipper = list(zip(splist, srlist, zip_eps.flatten()))  # Zip them up
            pack = np.array([self.get_intercept_params(x[0], x[1], x[2]) for x in zipper], dtype=object)  # Get param pack
            keys = ['epoch', 'lon', 'lat', 'theta', 'phi', 'ra', 'dec', 'sun_flag', 'moon_flag']
            vals = [np.degrees(pack[:, x].astype(float).reshape(len(epochs), self.rad_number).T) for x in range(6)]  # Make obs dict
            vals2 = [pack[:, x].astype(bool).reshape(len(epochs), self.rad_number).T for x in [6, 7]]
        
            vals = [zip_eps.T] + vals + vals2
            obs_dict = dict(zip(keys, vals))

            # obs_dict = {'epoch': epochs, 'lon': lons, 'lat': lats, 'theta': thetas, 'phi': phis, 
            #             'ra': ras, 'dec': decs, 'sun_flag': sun_flag, 'moon_flag': moon_flag}

            return obs_dict

    def make_antenna_grid(self, start_epoch, end_epoch, resolution, power_cutoff=0.5): 
        """ Similar to make_obs_grid, but each discrete observation contains an array of
            lat/lon/ra/dec values corresponding to rays sampling an antenna pattern

            This routine is still under construction, and will raise an error when called 
        """

        raise NotImplementedError('Still under construction')
        epochs = np.arange(start_epoch, end_epoch, resolution) 
        packets = np.zeros((self.rad_number, len(epochs)), dtype=object)

        ant_theta = self.ant_dict['theta']
        ant_phi = self.ant_dict['phi']
        
        for i, eps in enumerate(epochs): 
            for j, rads in range(self.rad_number): 
                lons = np.zeros((len(ant_theta), len(ant_phi)))
                lats = np.zeros((len(ant_theta), len(ant_phi)))
                thetas = np.zeros((len(ant_theta), len(ant_phi)))
                phis = np.zeros((len(ant_theta), len(ant_phi)))
                ras = np.zeros((len(ant_theta), len(ant_phi)))
                decs = np.zeros((len(ant_theta), len(ant_phi)))
                ref_ras = np.zeros((len(ant_theta), len(ant_phi)))
                ref_decs = np.zeros((len(ant_theta), len(ant_phi)))
                sun_flag = np.zeros((len(ant_theta), len(ant_phi)), dtype=bool)
                moon_flag = np.zeros((len(ant_theta), len(ant_phi)), dtype=bool)
                for k, th in enumerate(self.ant_dict['theta']): 
                    for l, phi in enumerate(self.ant_dict['phi']): 
                        if self.antenna_pattern(th, phi) >= power_cutoff: 
                            # Note, I'm not defining an azimuth convention yet, which
                            # will be an issue for non-symmetrical power patterns
                            dvec = spice.rotvec(np.array([1, 0, 0]), np.radians(th), 2)
                            dvec = spice.vhat(spice.rotvec(dvec, np.radians(phi), 1))

                            try: 
                                # Check for surface intercept
                                spoint, trgepc, srfvec = spice.sincpt('ELLIPSOID', 'EARTH', eps, 'ITRF93', 'NONE', 
                                                                      '{}'.format(self.naif_id), 
                                                                      'SPACECRAFT_{0}_RADIOMETER_{1}'.format(self.sc_number, j), 
                                                                      dvec)
                                lon, lat, theta, phi, ra, dec, found_sun, found_moon = self.get_intercept_params(spoint, srfvec, eps)

                                lons[k, l] = np.degrees(lon)
                                lats[k, l] = np.degrees(lat)
                                thetas[k, l] = np.degrees(theta)
                                phis[k, l] = np.degrees(phi)
                                ras[k, l] = np.nan
                                decs[k, l] = np.nan
                                ref_ras[k, l] = np.degrees(ra)
                                ref_decs[k, l] = np.degrees(dec)
                                sun_flag[k, l] = found_sun 
                                moon_flag[k, l] = found_moon 

                            except spice.utils.exceptions.NotFoundError: 
                                # Cant find surface intercept, get CK pointing 
                                rad_id = int(self.naif_id * 1e3 - 1 - j)
                                encep = spice.sce2t(eps)
                                cmat, encep = spice.ckgp(rad_id, encep, 0, 'J2000')
                                inertial_pointing = cmat.T @ dvec
                                ran, ra, dec = spice.recrad(inertial_pointing)
                                found_sun, found_moon = self.check_moon_sun(eps, inertial_pointing)

                                lons[k, l] = np.nan
                                lats[k, l] = np.nan
                                thetas[k, l] = np.nan
                                phis[k, l] = np.nan
                                ras[k, l] = np.degrees(ra)
                                decs[k, l] = np.degrees(dec)
                                ref_ras[k, l] = np.nan
                                ref_decs[k, l] = np.nan
                                sun_flag[k, l] = found_sun 
                                moon_flag[k, l] = found_moon 

    def make_parallel_grid(self, start_epoch, end_epoch, resolution, bounds=None, ndiv=4, pool=None, nproc=4): 
        """ Parallel implementation of make_obs_grid, which computes the grid over subwindows
            using spacecraft copies  

            :param start_epoch: Observation start epoch
            :param end_epoch: Observation end epoch
            :param resolution: Time resolution of tracks
            :param bounds: Tuple of tuples ((lower_lat, upper_lat), (lower_lon, upper_lon))
                           Default is None
            :param ndiv: Number of time blocks to feed to the pool (generally the same as ndiv)
            :param pool: Process pool. If not assigned, this function creates one 
            :param nproc: Number of processes in the pool, overriden if pool is provided

            :return: A dictionary of epochs, longitudes, latitudes, incidence elevations (theta), 
                    incidence azimuths (phi), right ascension and declination in the direction of 
                    specular reflection, and flags for if the sun and moon are in that direction 

                    The shape of the arrays in each dictionary value (except epochs) is NxM where N is the number 
                    of radiometers on the spacecraft, and M is the length of the epochs value

        """

        epochs = np.arange(start_epoch, end_epoch, resolution)
        epoch_list = np.array_split(epochs, ndiv)

        if not pool: 
            local_pool = True 
            pool = Pool(processes=nproc)
        else: 
            local_pool = False

        # Build craft copy queue
        craft_list = [] 
        for n in range(nproc): 
            craft_list.append(copy.deepcopy(self))
        craft_list.extend(craft_list * (ndiv - 1))

        zipper = list(zip(craft_list, epoch_list, repeat(bounds), range(ndiv)))
        dicts = pool.starmap(self._mpg, zipper)

        if local_pool: 
            pool.close()
            pool.join()

        # Now that's a comprehension!
        # obs_dict = {k: np.concatenate([d.get(k) for d in dicts]) for k in {k for d in dicts for k in d}}
        keys = ['epoch', 'lon', 'lat', 'theta', 'phi', 'ra', 'dec', 'sun_flag', 'moon_flag']
        dicts = [d for d in dicts if d]  # Filter out empty dicts
        obs_dict = {k: np.concatenate([d.get(k) for d in dicts], axis=1) for k in keys}
        return obs_dict

    @staticmethod
    def _mpg(craft, eps, bounds, i=0):
        """ Parallel gridder map function 
        """
        craft.furnish()
        if bounds is None: 
            obs_dict = craft._mog(eps, i)
        else: 
            obs_dict = craft._mog_bound(eps, bounds, i)
        craft.unload()
        return obs_dict

    def get_intercept_params(self, spoint, srfvec, epoch): 
        # Get lat and lon 
        lon, lat, tall = spice.recgeo(spoint, 6378.137, 1 / 298.257223563)  # WGS84 ellipsoid

        # Get theta and phi incidence angle

        # Slightly slower 
        # colat = np.pi / 2 - lat 
        # m1 = spice.rotate(np.pi / 2, 3)
        # m2 = spice.rotate(colat, 2)
        # m3 = spice.rotate(lon, 3)
        # ef2topo = m1 @ m2 @ m3 

        # Slightly faster
        sinl = np.sin(lon)
        cosl = np.cos(lon)
        sinp = np.sin(lat)
        cosp = np.cos(lat)
        ef2topo = np.array([[-sinl, cosl, 0],
                            [-cosl * sinp, -sinl * sinp, cosp],
                            [cosl * cosp, sinl * cosp, sinp]])
        
        topovec = ef2topo @ srfvec
        theta = np.pi - np.arccos(topovec[2] / spice.vnorm(topovec))
        phi = np.pi + np.arctan2(topovec[1], topovec[0])

        # Get specular ra and dec 
        ra, dec, point_vec = self.specular_reflection(spoint, srfvec)

        # Get sun and moon flags 
        found_sun, found_moon = self.check_moon_sun(epoch, point_vec)

        return lon, lat, theta, phi, ra, dec, found_sun, found_moon

    def get_manual_elems(self, ndt2=0., ndd6=0., bstar=0., inclination=0., raan=0., eccentricity=0.,
                         argperigee=0., meananomaly=0., height=500000., tle_epoch=0.): 
        """ A helper class that forms an element list from TLE inputs. Use this function also to reference
            the units for each parameter required for successful TLE kernel creation 
 
            :param ndt2: First derivative of mean motion in radians/minute^2
            :param ndd6: Second derivative of mean motion in radians/minute^3
            :param bstar: Radiation pressure coefficient (unitless)
            :param inclination: Inclination in radians 
            :param raan: Right Ascension of Ascending Node in radians 
            :param eccentricity: Eccentricity (unitless)
            :param argperigee: Argument of Perigee in radians 
            :param meananomaly: Mean anomaly in radians 
            :param height: Height of semimajor axis above the surface of Earth in meters
            :param tle_epoch: Epoch specified in the TLE set in seconds past J2000

            :return: A list of elements that can be used as an argument to spacecraft.write_tle_kernels

        """
        semimaj = height + self.earth_mean_radius
        meanmotion = np.sqrt(self.mu / semimaj ** 3)  # in radians/sec
        elems = [ndt2, ndd6, bstar, inclination, raan, eccentricity, argperigee, meananomaly, meanmotion * 60, tle_epoch]
        return elems

    @staticmethod
    def read_tle(filename): 
        """ Reads TLE from file using the SPICE getelm function 

            :param filename: TLE text file

            :returns: 
                - epoch - SPICE formatted epoch
                - elements - SPICE formatted elements
        """
        f = open(filename, 'r')
        g = f.readlines()
        h = [g[x].strip() for x in range(len(g))]
        epoch, elements = spice.getelm(2000, 500, h)
        f.close()
        return epoch, elements 

    def write_tle_file(self, epoch=datetime(2000, 1, 1), inclination=0, raan=0, eccentricity=0, argperigee=0, meananomaly=0, height=0, revnumber=0):
        """ Writes two line element strings from inputs 
            All arguments should be floats except for epoch, which should be a datetime object
            Hardcoding the derivatives of mean motion and the radiation pressure drag to zero

            :param inclination: Inclination in radians 
            :param raan: Right Ascension of Ascending Node in radians 
            :param eccentricity: Eccentricity (unitless)
            :param argperigee: Argument of Perigee in radians 
            :param meananomaly: Mean anomaly in radians 
            :param height: Height of semimajor axis above the surface of Earth in meters
            :param revnumber: Number of revolutions around the Earth at the desired epoch

        """  
        if type(epoch) is not datetime: 
            raise ValueError("epoch is not a datetime object")

        semimaj = height + self.earth_mean_radius
        meanmotion = np.sqrt(self.mu / semimaj ** 3) * 60  # in radians/min

        satname = "FOAMSAT    "  # 11 characters

        # Line 1 
        ref_year = datetime(epoch.year, 1, 1)
        days = (epoch - ref_year).total_seconds() / 84600
        epoch_string = "{0}{1:012.8f}".format(epoch.strftime('%y'), days)
        line1 = "1 00000U 000000 {0} 0.00000000 00000-0 00000-0 0 ".format(epoch_string)
        line1 += self.tle_checksum(line1)

        # Line 2 
        ecc = int("{0:.7f}".format(eccentricity).split('.')[1])
        line2 = "2 00000 {0:07.4f} {1:08.4f} {2} {3:08.4f} {4:08.4f} {5:011.8f} {6:05d}".format(inclination, raan, ecc, argperigee, meananomaly, meanmotion, revnumber)
        line2 += self.tle_checksum(line2)
        # Demo TLE, see space-track.org and wiki for formatting
        self.tle = satname + '\n' + line1 + '\n' + line2

    @staticmethod
    def tle_checksum(string): 
        checksum = 0 
        for i in string:
            if i.isdigit(): 
                checksum += int(i)
            elif i == '-':
                checksum += 1
        return str(checksum % 10)

    @staticmethod
    def specular_reflection(spoint, srfvec):
        """ Calculates the specular reflection vector for a given spacecraft look direction 

            :param spoint: Surface intercept point on target body
            :param srfvec: Vector from observer to intercept point 

            :return: RA, Dec in radians and unit vector in specular reflection direction
        
        """

        # Old, expensive way of doing it
        # angle = np.pi - spice.vsep(spoint, srfvec)
        # normal = spice.vcrss(spoint, srfvec)
        # rotframe = spice.twovec(normal, 1, spoint, 2)
        # invframe = spice.invert(rotframe)
        # flipvec = spice.vrotv(rotframe @ srfvec, np.array([1, 0, 0]), np.array([np.pi]))
        # rotvec = spice.vrotv(flipvec, np.array([1, 0, 0]), 2 * angle)
        # point_vec = invframe @ rotvec

        # Cheaper way 
        n = spice.vhat(spoint)
        d = spice.vhat(srfvec)
        point_vec = d - 2 * np.dot(d, n) * n

        ran, ra, dec = spice.recrad(point_vec)

        return ra, dec, point_vec

    @staticmethod
    def check_moon_sun(epoch, point_vec):
        """ Check if the moon and the sun intersect with a pointing vector. 
            This can be used with the outputs of the specular_reflection function 

        """

        # Spiceypy raises errors if intercepts aren't found, so found flags have to deal with this
        try: 
            spoint, trgepc, srfvec = spice.sincpt('ELLIPSOID', 'SUN', epoch, 'IAU_SUN', 'LT+S', 'EARTH', 'ITRF93', point_vec)
            found_sun = True 
        except spice.utils.exceptions.NotFoundError: 
            found_sun = False 
        try: 
            spoint, trgepc, srfvec = spice.sincpt('ELLIPSOID', 'MOON', epoch, 'MOON_ME', 'LT+S', 'EARTH', 'ITRF93', point_vec)
            found_moon = True
        except spice.utils.exceptions.NotFoundError: 
            found_moon = False 
        return found_sun, found_moon


# Spacecraft-related utility methods 

def make_smap(start_epoch, end_epoch, epoch_res=1, sc_number=0): 
    """ Builds the SMAP spacecraft using the spacecraft module.

        Spin platform rotation rate: 14.6 rpm 

        Antenna look angle (emission angle): 35.5 deg. (40 deg.)

        Orbital altitude: 670 km 

        Integration time: 17.5 ms (resolution provided for level 1 products)
    """ 
    manual_furnish()
    smap_craft = spacecraft(sc_number=sc_number)
    tle = ['1 40376U 15003A   20247.33652497  .00000025  00000-0  13841-4 0  9991',
           '2 40376  98.1240 252.7233 0001762  90.8135 269.3273 14.63327320298611']
    epoch, elements = spice.getelm(2000, 500, tle)
    smap_craft.write_tle_kernels(elems=elements, tle_epoch=epoch, start_epoch=start_epoch, 
                                 end_epoch=end_epoch, epoch_res=epoch_res)
    # look_angle, look_axis, rpm, scan_axis
    smap_craft.write_radiometer_ck(35.5, 'Y', 14.6, 'X')

    return smap_craft


def make_aquarius(start_epoch, end_epoch, epoch_res=1, sc_number=0): 
    """ Builds the Aquarius spacecraft using the spacecraft module.

        Horn angles: 25.8 deg., 33.8 deg., 40.3 deg.

        Orbital altitude: 657 km 

        Integration time: 12s (resolution provided for level 1 products)
    """ 
    
    # From an RSS memo
    # Aquarius horns (pre-launch, 0.5 degrees accurate) were pointed as 
    # Horn      Elevation       Azimuth
    # Horn 1    25.82           9.84
    # Horn 2    33.82           -15.29
    # Horn 3    40.37           6.55
    # The total angle for each radiometer can then be computed as follows
    # horn = [1, 2, 3]
    # nadir_angle = [-25.82, -33.82, -40.37]
    # azimuth_angle = [9.84, -15.29, 6.55]
    # for i in range(len(horn)):  
    #     rm = spice.rotate(np.radians(azimuth_angle[i]), 3)
    #     rm2 = spice.rotate(np.radians(nadir_angle[i]), 2)
    #     axis, angle = spice.raxisa(rm @ rm2)
    #     print('Horn {0}: Axis {1}, Angle {2}'.format(horn[i], axis, np.degrees(angle)))
    manual_furnish()
    aq_craft = spacecraft(sc_number=sc_number)
    tle = ['1 37673U 11024A   21237.59766439  .00000119  00000-0  26113-4 0  9996',
           '2 37673  98.0067 245.2609 0001526  59.7629  52.4724 14.73133816548758']
    epoch, elements = spice.getelm(2000, 500, tle)
    aq_craft.write_tle_kernels(elems=elements, tle_epoch=epoch, start_epoch=start_epoch, 
                               end_epoch=end_epoch, epoch_res=epoch_res)
    axis = np.array([0.08032594, 0.93313411, -0.35044041])
    aq_craft.write_radiometer_ck(27.60, axis, 0., 'X')
    axis = np.array([-0.12187662, 0.90798238, 0.40089161])
    aq_craft.write_radiometer_ck(37.02, axis, 0., 'X')
    axis = np.array([0.05645089, 0.98652661, -0.15355307])
    aq_craft.write_radiometer_ck(40.88, axis, 0., 'X')

    return aq_craft


def track_viewer_2D(craft, start_epoch, end_epoch, resolution): 
    """ Plots spacecraft and radiometer look tracks on a 2D map

        :param craft: A spacecraft obect 
        :param start_epoch: Starting epoch in seconds 
        :param end_epoch: Ending epoch in seconds 
        :param resolution: Time resolution in seconds 
    
    """ 
    track_dict = craft.make_track_grid(start_epoch, end_epoch, resolution)
    lons = track_dict['lon']
    lats = track_dict['lat']
    plt.figure()
    ax = plt.axes(projection=ccrs.PlateCarree())
    ax.coastlines()
    ax.plot(lons, lats, transform=ccrs.PlateCarree(), linestyle='none', marker='.', color='k', 
            label='Spacecraft tracks', ms=2)
    obs_dict = craft.make_obs_grid(start_epoch, end_epoch, resolution)
    lons = obs_dict['lon']
    lats = obs_dict['lat']
    shape = np.shape(lons)
    for i in range(shape[0]): 
        ax.plot(lons[i, :], lats[i, :], transform=ccrs.PlateCarree(), linestyle='none', marker='.',
                label='Radiometer {}'.format(i), ms=2)
    ax.set_extent([-180, 180, -90, 90])


def track_viewer_animated(craft, start_epoch, end_epoch, resolution, saveplots=True): 
    """ Animated plots spacecraft and radiometer look tracks on a 2D map 
        overtime with a 3D plot of observing geometry

        :param craft: A spacecraft obect 
        :param start_epoch: Starting epoch in seconds 
        :param end_epoch: Ending epoch in seconds 
        :param resolution: Time resolution in seconds 

    """ 
    
    fig = plt.figure(figsize=(20, 10)) 
    ax = fig.add_subplot(121, projection='3d')
    ax2 = fig.add_subplot(122, projection=ccrs.PlateCarree())
    ax2.coastlines()

    # Plot Earth 
    u = np.linspace(0, 2 * np.pi, 500)
    v = np.linspace(0, np.pi, 500)
    rad = craft.earth_mean_radius / 1e3
    x = rad * np.outer(np.cos(u), np.sin(v))
    y = rad * np.outer(np.sin(u), np.sin(v))
    z = rad * np.outer(np.ones(np.size(u)), np.cos(v))
    ax.plot_wireframe(x, y, z, color='black', alpha=0.1)

    epochs = np.arange(start_epoch, end_epoch, resolution)
    obs_dict = craft.make_obs_grid(start_epoch, end_epoch, resolution)
    epochs = obs_dict['epoch']

    for t in range(len(epochs)):
        pos, times = spice.spkpos('{}'.format(craft.naif_id), epochs[t], 'ITRF93', 'NONE', '399')
        ax.scatter(pos[0], pos[1], pos[2], color='blue')
        pos2, times = spice.spkpos('399', epochs[t], 'SPACECRAFT_{}'.format(craft.sc_number), 'NONE', 
                                   '{}'.format(craft.naif_id))
        mat = spice.pxform('SPACECRAFT_{}'.format(craft.sc_number), 'ITRF93', epochs[t])
        pos2 = mat @ pos2
        lon, lat, tall = spice.recgeo(pos, 6378.137, 1 / 298.257223563)  # WGS84 ellipsoid
        height = tall
        pos2_norm = pos2 * (height / np.linalg.norm(pos))
        ax.quiver(pos[0], pos[1], pos[2], pos2_norm[0], pos2_norm[1], pos2_norm[2], color='red')
        for i in range(craft.rad_number): 
            try:
                spoint, trgepc, srfvec = spice.sincpt('ELLIPSOID', 'EARTH', epochs[t], 'ITRF93', 'NONE', 
                                                      '{}'.format(craft.naif_id), 
                                                      'RADIOMETER_{}'.format(i), np.array([1, 0, 0]))
                ax.scatter(spoint[0], spoint[1], spoint[2], color='green')
                ll_arr2 = spice.reclat(spoint)
                # print("Lon/Lat SPOINT: %f, %f" % (np.degrees(ll_arr2[1]), np.degrees(ll_arr2[2])))
                ax.quiver(spoint[0], spoint[1], spoint[2], -srfvec[0], -srfvec[1], -srfvec[2], color='purple')

            except spice.utils.exceptions.NotFoundError: 
                ll_arr2 = np.array([1, 0, 0])

            enctime = spice.sce2c(craft.naif_id, epochs[t])
            cmat, clkout = spice.ckgp(int(craft.naif_id * 1e3 - 1 - i), enctime, 0, 'ITRF93')
            cpoint = cmat.T @ (np.array([1, 0, 0]))
            cpoint_norm = 2 * cpoint * (height / np.linalg.norm(cpoint))
            ax.quiver(pos[0], pos[1], pos[2], cpoint_norm[0], cpoint_norm[1], cpoint_norm[2], color='g')
            cpoint = cmat.T @ (np.array([0, 1, 0]))
            cpoint_norm = 2 * cpoint * (height / np.linalg.norm(cpoint))
            ax.quiver(pos[0], pos[1], pos[2], cpoint_norm[0], cpoint_norm[1], cpoint_norm[2], color='m')
            cpoint = cmat.T @ (np.array([0, 0, 1]))
            cpoint_norm = 2 * cpoint * (height / np.linalg.norm(cpoint))
            ax.quiver(pos[0], pos[1], pos[2], cpoint_norm[0], cpoint_norm[1], cpoint_norm[2], color='y')

            ll_arr = spice.reclat(pos)
            ax2.plot(np.degrees(ll_arr[1]), np.degrees(ll_arr[2]), transform=ccrs.PlateCarree(), linestyle='none', marker='.', color='blue')
            ax2.plot(np.degrees(ll_arr2[1]), np.degrees(ll_arr2[2]), transform=ccrs.PlateCarree(), linestyle='none', marker='.', color='green')
            ax2.set_extent([-180, 180, -90, 90])
            # print("Lon/Lat SUB: %f, %f" % (np.degrees(ll_arr[1]), np.degrees(ll_arr[2])))

        if t == 0:
            # plt.show(block=False)
            if os.path.exists('animated_tracks'): 
                os.rmdir('animated_tracks')
            os.mkdir('animated_tracks')
        else:
            # plt.draw()
            pass
        plt.tight_layout()
        if saveplots: plt.savefig('animated_tracks/chart_{}'.format(t), dpi=90)
        # plt.pause(0.2)


def revisit_time(craft, start_epoch, end_epoch, epoch_res, 
                 grid_mode='linear', grid_res=0.25, grid_stop=1, 
                 plots=False, parallel=False, **kwargs):
    
    """ Computes revisit counts for a spacecraft over a given time range 

        :param craft: Spacecraft object
        :param start_epoch: Starting epoch in seconds since J2000
        :param end_epoch: Ending epoch in seconds since J2000
        :param epoch_res: Time sample resolution in seconds 
        :param grid_mode: Options for grid mode are: 

                            * 'linear' (default): Uniform lat/lon grid resolution set by grid_res
                            
                            * 'ease': Scaled EASE grid with the finest grid resolution set by grid_res
                            
                            * 'cosine': Uniform longitude spacing (grid_res) and a cosine-graded latitude spacing ranging from grid_res to grid_stop. Not exact! 

        :param grid_res: See above, in degrees
        :param grid_stop: See above, in degrees. Only used for grid_mode='cosine'
        :param plots: If True, plot results as pcolormesh
        :param parallel: If true, uses craft.make_parallel_grid (and associated 
                         kwargs), to generate observations in parallel
    """ 

    if parallel: 
        obs_dict = craft.make_parallel_grid(start_epoch, end_epoch, epoch_res, **kwargs)
    else: 
        obs_dict = craft.make_obs_grid(start_epoch, end_epoch, epoch_res, **kwargs)
    obs_dict = {key: value.ravel() for key, value in obs_dict.items()}
    
    if grid_mode == 'linear':
        lat_bins = np.arange(-90, 90 + grid_res, grid_res)
        lon_bins = np.arange(-180, 180 + grid_res, grid_res)
    elif grid_mode == 'ease': 
        dim = (4872, 11568)
        land_lat = np.fromfile(os.path.join(cache_path, 'landmask', 'EASE2_M03km.lats.11568x4872x1.double'), dtype=np.float64)
        lat_bins = land_lat.reshape(dim)[::-1, 0]
        land_lon = np.fromfile(os.path.join(cache_path, 'landmask', 'EASE2_M03km.lons.11568x4872x1.double'), dtype=np.float64)
        lon_bins = land_lon.reshape(dim)[0]
        for i in range(20):
            lat_bins = np.insert(lat_bins, 0, lat_bins[0] - 0.25548359)
            lat_bins = np.append(lat_bins, lat_bins[-1] + 0.25548359)
        lat_bins[0] = -90
        lat_bins[-1] = 90
        sample = int(np.round(grid_res / (360 / dim[-1])))
        lat_bins = lat_bins[::sample]
        lon_bins = lon_bins[::sample]
        lat_bins[0] = -90
        lat_bins[-1] = 90
        lon_bins[0] = -180
        lon_bins[-1] = 180

    elif grid_mode == 'cosine':
        lat_bins = _cosine_grid(0, 90, grid_res, grid_stop)
        lat_bins = np.concatenate([np.flip(-lat_bins[1:]), lat_bins])
        lon_bins = np.arange(-180, 180 + grid_res, grid_res)
    else:
        raise ValueError("Grid mode must be 'linear', 'ease', or 'cosine'")
    
    llat, llon = np.meshgrid(lat_bins, lon_bins)
    grid = np.zeros((len(lat_bins) - 1, len(lon_bins) - 1))
    ep_mod = (obs_dict['epoch'] - obs_dict['epoch'][0]) % craft.orbital_period
    dexs = np.where(np.diff(ep_mod) < 0)  # Find zero points, or new orbits
    dexs = np.concatenate([np.zeros(1, dtype=int), dexs[0], [len(ep_mod)]])
    
    # Old method, clean but double-counts 
    # for i in range(1, len(dexs)): 
    #     lats = obs_dict['lat'][dexs[i - 1]:dexs[i]]
    #     lons = obs_dict['lon'][dexs[i - 1]:dexs[i]]
    #     pack = np.histogram2d(lats, lons, bins=[lat_bins, lon_bins])
    #     cg = pack[0].astype(bool).astype(int)
    #     grid += cg

    # New method, using pandas 
    nset = pd.DataFrame.from_dict(obs_dict).iloc[:, :3]
    nset['epoch'] = nset['epoch'] - nset['epoch'][0]
    # Binning 
    lat_dex = np.digitize(nset['lat'], lat_bins)
    lon_dex = np.digitize(nset['lon'], lon_bins)
    nset['lat'] = lat_bins[lat_dex - 1]
    nset['lon'] = lon_bins[lon_dex - 1]

    # Time filtering critera
    # Removes values in bins within less than half 
    # of an orbital period of each other
    # prepend ensures first value is kept
    def func(x):
        z = x['epoch']
        if len(z) > 1:
            return z[np.diff(z, prepend=-1e10) > (craft.orbital_period / 2)]
        else:
            return z

    nsgb = nset.groupby(['lat', 'lon'])
    nsout = nsgb.apply(func)  # Filters out lat and lon
    nsout = nsout.astype(bool).astype(int)
    nsout = nsout.groupby(level=[0, 1]).sum()

    lat_vals = nsout.index.get_level_values(0) 
    lon_vals = nsout.index.get_level_values(1)
    grid, *_ = np.histogram2d(lat_vals, lon_vals, bins=[lat_bins, lon_bins], weights=nsout.values)

    if plots: 
        fig = plt.figure()
        ax = fig.add_subplot(111, projection=ccrs.PlateCarree())
        ax.coastlines()
        im = ax.pcolormesh(lon_bins, lat_bins, grid, shading='auto', alpha=0.8, cmap='turbo')
        fig.colorbar(im, fraction=0.03, pad=0.04, label='Revisits')
    return grid, lon_bins, lat_bins, obs_dict


def angle_conversion(h, in_angle, in_angle_type='incidence', R=6371): 
    """ Utility to convert antenna look angle to Earth surface incidence angle 
        or vice versa. Assumes either angle is < 90 deg

        :param h: Height above surface in km 
        :param in_angle: Input angle in degrees
        :param in_angle_type: Either surface 'incidence' or antenna 'look'
        :param R: Radius of body in km (defaults to Earth, 6371)
    """ 
    in_angle = np.radians(in_angle)
    if in_angle_type == 'incidence': 
        out = np.arcsin(R / (R + h) * np.sin(np.pi - in_angle))
    elif in_angle_type == 'look': 
        out = np.arcsin((R + h) / R * np.sin(in_angle))
    else: 
        raise ValueError('in_angle_type must be either "incidence" or "look"')

    return np.degrees(out)


def _cosine_grid(start, stop, dx_start, dx_stop): 
    """ Ad hoc grid spacing method which, when given a start and stop point, 
        generates a sequence of numbers x with properties diff(x[:2]) == dx_start
        and diff(x[-2:]) == dx_stop with a cosine taper. 
        This is an iterative, inexact method which 
        is useful for generating scaled latitude grids for visualization
    """ 

    if (dx_start > 25) or (dx_stop > 25): 
        raise ValueError('Grid too coarse for this method')
    elif dx_start > dx_stop: 
        raise ValueError('dx_start must be smaller than dx_stop')

    a = dx_stop 
    b = dx_stop - dx_start
    n = 3
    x = np.linspace(0, n, n + 1)
    y = a - b * np.cos(np.pi / 2 * x / (n + 1))
    csy = np.cumsum(y)
    y = start + csy / csy[-1] * (stop - start)
    y = np.insert(y, 0, start)
    dy = np.diff(y)

    start_check = abs(dy[0] - dx_start)
    stop_check = abs(dy[-1] - dx_stop)

    going = False
    while True: 
        n += 1
        if n == 1e4: break
        x = np.linspace(0, n, n + 1)
        y = a - b * np.cos(np.pi / 2 * x / (n + 1))
        csy = np.cumsum(y)
        y = start + csy / csy[-1] * (stop - start)
        y = np.insert(y, 0, start)
        dy = np.diff(y)        
        strc = abs(dy[0] - dx_start)
        stpc = abs(dy[-1] - dx_stop)

        if (strc < start_check) and (stpc < stop_check): 
            start_check = strc 
            stop_check = stpc 
            going = True
        elif going:
            break

    x = np.linspace(0, n, n + 1)
    y = a - b * np.cos(np.pi / 2 * x / (n + 1))
    csy = np.cumsum(y)
    y = start + csy / csy[-1] * (stop - start)
    y = np.insert(y, 0, start)

    return y


def strings_to_epochs(*args): 
    """ Converts strings to epochs 
        useable by spacecraft generator objects.

        Example string format - '2020 MAY 24 9:00 UTC'
    """ 

    spice.furnsh(os.path.join(cache_path, 'spice/latest_leapseconds.tls'))
    output = tuple(spice.str2et(arg) for arg in args)
    spice.unload(os.path.join(cache_path, 'spice/latest_leapseconds.tls'))
    return output


def merge_obs_dict(*args): 
    """ Merge any number of observing dictionaries into one 
    """ 
    obs_list = list(args)
    keys = args[0].keys()
    obs_list = [d for d in obs_list if d]  # Filter out empty dicts
    obs_dict = {k: np.concatenate([d.get(k) for d in obs_list], axis=0) for k in keys}
    return obs_dict










    



import os 
import numpy as np
from numpy import sin, cos, radians
import pandas as pd 
import scipy.interpolate as spi 
import scipy.constants as spc 
from cftime import date2num

import foam.geomag as geomag

from .utils.config import cache_path, dir_path
from .utils import reader

try:
    import iri2016
    has_IRI = True
except ImportError: 
    has_IRI = False


class ionosphere(): 
    """ The ionosphere class reads maps of Total Electron Content (TEC) from CDDIS ancillary files
        or simulates these maps using the IRI2016 package
        
        :param datetime: Either a single date or a pair of dates bracketing an interval of time
                         Several formats are permissible, such as
                         - String or iterable of strings
                         - Python datetime or iterable of datetimes 
                         - Numpy datetime64 or iterable
                         - Pandas timestamp or iterable 

        :param online: If true, ionospheric data is downloaded from CDDIS servers based on the datetime parameter.
        :param tec_file: Filename for a CDDIS ionosphere TEC map. If None, the FOAM cache default is used
        :param mag_file: Filename for World Magnetic Model coefficients, If None, the FOAM cache default is used 
        :param tec_reader: Reader object for TEC, default is IONEXReader
        :param IRI: If true, ionospheric TEC is simulated using the IRI2016 package. 
                    This mode is very experimental, and subject to be removed
        :param verbose: If true, prints verbose output

    """

    time_reference = 'seconds since 2000-01-01 12:00:0.0'  # J2000

    def __init__(self, datetime='2015-01-01', online=False, tec_file=None, mag_file=None, 
                 tec_reader=reader.IONEXReader, tec_reader_kwargs=None, IRI=False, verbose=False, **kwargs):

        self.datetime = pd.to_datetime(datetime)
        self.online = online 
        self.tec_file = tec_file
        self.mag_file = mag_file
        self.tec_reader = tec_reader
        if tec_reader_kwargs is None: 
            self.tec_reader_kwargs = {}
        else:    
            self.tec_reader_kwargs = tec_reader_kwargs
        self.IRI = IRI
        self.verbose = verbose

        for key, value in kwargs.items(): 
            setattr(self, key, value)

        if self.tec_file is None and not self.online:
            self.datetime = pd.to_datetime(['2015-01-01 12:00'])
            self.tec_file = os.path.join(cache_path, 'ionosphere', 'jplg0010.05i')
        if self.mag_file is None: 
            self.mag_file = os.path.join(dir_path, 'assets', 'magneticfield', 'WMM2020.COF')

        if self.IRI: 
            if not has_IRI: 
                raise RuntimeError('User needs to install the iri2016 package to use this mode')
            self.make_ionosphere()
        else: 
            self.read_ionosphere()

    def read_ionosphere(self):
        """ Reads ionospheric state 
        """ 
        self.tec_reader = self.tec_reader(self.datetime, self.online, file=self.tec_file)
        self.TEC_interp = self.tec_reader.read(self.tec_reader_kwargs)
        
    def make_ionosphere(self): 
        """ Uses the IRI2016 implementation of the International Reference Ionosphere (http://irimodel.org/)
            to generate maps of ionospheric TEC. Due to the slow speed of IRI runs, this is only done for a single time
        """
        print('Using IRI 2016 to calculate ionosphere TEC on a 5 degree grid')
        print('This can take several minutes')

        # 5 degree grid
        grid_lat = np.linspace(90, -90, 36)
        grid_lon = np.linspace(-180, 180, 72)
        TEC = np.array([iri2016.IRI(self.datetime[0], [0, 800, 800], x, y).TEC.values for x in grid_lat for y in grid_lon])
        TEC = TEC.reshape(36, 72) / 1e16  # Convert to TECu
        self.TEC_interp = spi.RegularGridInterpolator((grid_lat[::-1], grid_lon), TEC[::-1, :], bounds_error=False, fill_value=18)

    @staticmethod
    def faraday_rotation(far_angle, TB): 
        """ Applies faraday rotation to the polarized brightness temperatures
            
            :param far_angle: Angle of rotation in radians
            :param TB: Block matrix of polarimetric brightness temperatures [TBV, TBH, U, V] (Size 4xMx...)

            :return: 
        """

        R = np.array([[cos(far_angle)**2, sin(far_angle)**2, 0.5 * sin(2 * far_angle), np.zeros(np.shape(far_angle))],
                    [sin(far_angle)**2, cos(far_angle)**2, -0.5 * sin(2 * far_angle), np.zeros(np.shape(far_angle))],
                    [-sin(2 * far_angle), sin(2 * far_angle), cos(2 * far_angle), np.zeros(np.shape(far_angle))], 
                    [np.zeros(np.shape(far_angle)), np.zeros(np.shape(far_angle)), np.zeros(np.shape(far_angle)), np.ones(np.shape(far_angle))]])
        
        # Correct nans 
        mask = np.isnan(TB)
        TB[mask] = 0 
        TA = np.einsum('ij...,j...->i...', R, TB)
        TA[mask] = np.nan
        return TA

    def compute_faraday_angle(self, frequency, time, lat, lon, theta, phi, in_epoch=False, use_time=True):
        """ Computes the angle of ionospheric faraday rotation for polarized microwave emission.
            This calculation uses a Python adaptation of the World Magnetic Model written by Christopher Weiss 
            https://github.com/cmweiss/geomag. Due to its relatively small size, the geomag package has been 
            included directly with FOAM without requiring a download. It has also been modified to support numpy arrays. 
            
            :param frequency: Measurement frequency in MHz (shape N)
            :param time: String or array of times, converted to interpolator
                         epoch reference if in_epoch=False
            :param lat: Latitude in degrees 
            :param lon: Longitude in degrees 
            :param theta: Radiometer elevation angle in degrees
            :param phi: Radiometer azimuth angle in degrees 
            :param in_epoch: See above
            :param use_time: Toggles use of time in case of heterogeneous ancillary data

            :return: Faraday rotation angle in radians (shape Nx...)
        """ 

        if not use_time: 
            time = self.datetime[0]
            in_epoch = False
        if not in_epoch: 
            time = date2num(pd.to_datetime(time).to_pydatetime(), self.time_reference)

        if self.IRI: 
            TEC = self.TEC_interp((lat, lon))
        else: 
            TEC = self.TEC_interp((time, lat, lon))
        gm = geomag.GeoMag(self.mag_file)
        mag = gm.GeoMag(lat, lon, h=(675e3 / spc.foot))
        x = sin(radians(theta)) * cos(radians(phi))
        y = sin(radians(theta)) * sin(radians(phi))
        z = cos(radians(theta)) 
        mag_field = -(mag.bx * x + mag.by * y + mag.bz * z)
        frequency = np.squeeze(frequency)
        ghz = frequency / 1e3
        TEC = TEC[..., np.newaxis]
        mag_field = mag_field[..., np.newaxis]
        far_angle = np.radians(1.3549e-5 / ghz**2 * TEC * mag_field)
        far_angle = np.moveaxis(far_angle, -1, 0)
        return far_angle



import os 
import pickle
import h5py 
import numpy as np
import scipy.interpolate as spi 
import scipy.constants as spc

from foam.utils.config import dir_path, cache_path


class sky(): 
    """ The sky class implements models of all contributions to measured microwave emission above the ionosphere. This includes 
        the sun, moon, continuum galactic emission, and cosmic microwave background. See documentation for a description of these 
        models 

        :param scattered_galaxy: Toggles between specular reflection (False) and ocean surface scattering (True). Default is True 
        :param galaxy_file: Input galaxy map as a function of right ascension and declintion. If None, the FOAM default is used.
                            Currently, only the FOAM default is supported, but this may be changed in the future
        :param verbose: Toggles verbose output

    """ 

    galaxy_file = os.path.join(cache_path, 'galaxy/TBSkyLbandAquarius.h5')
    scattered_galaxy_file = os.path.join(dir_path, 'assets/galaxy/galaxy_interpolator_1deg.p')

    def __init__(self, scattered_galaxy=True, galaxy_file=None, verbose=False): 
        self.scattered_galaxy = scattered_galaxy
        if galaxy_file is not None: 
            self.galaxy_file = galaxy_file 
        elif self.scattered_galaxy: 
            self.galaxy_file = self.scattered_galaxy_file    
        self.verbose = verbose

        self.read_galaxy()

    def read_galaxy(self): 
        """ Reads galaxy map as a function of right acension vs declination and creates an interpolator
            If ocean scattering is specified, the interpolator is loaded from a pre-defined pickle. 

        """

        if self.scattered_galaxy: 
            if self.verbose: print('Using approximate galaxy reflection from wind-roughened ocean')
            self.galaxy_interp = pickle.load(open(self.galaxy_file, 'rb'))[0]
        else: 
            if self.verbose:
                print('Using specular galaxy map')
                print('Galaxy map file: %s' % self.galaxy_file)
            gal_map = h5py.File(self.galaxy_file, 'r')  # Cell size is 9.8572e-6 steradians
            self.galaxy_interp = spi.RegularGridInterpolator((gal_map['Right_Ascension'][:], gal_map['Declination'][:]), 
                                                        gal_map['TB_Cas_A_beam'][:, :] - 2.73, bounds_error=False, fill_value=0)  # Other options here include 'TB_Cas_A_1cell' and 'TB_no_Cas_A'
            gal_map.close()

        self.Tcmb = 2.73

    def galaxy_brightness(self, frequency, ra, dec, wind): 
        """ Retrieves galactic microwave brightness temperature from Dinnat et al. map
            
            :param frequency: Frequency in MHz (Size O)
            :param ra: Right ascension in degrees (Size MxN)
            :param dec: Declination in degrees (Size MxN)
            :param wind: 10-m wind speed in m/s (Size MxN)
            :return: Galactic brightness temperature (Size OxMxN)

        """ 

        frequency = frequency[:, np.newaxis]
        ra = ra[np.newaxis, :]
        dec = dec[np.newaxis, :]
        wind = wind[np.newaxis, :]
        if self.scattered_galaxy: 
            gal_Tb = self.galaxy_interp((wind, ra, dec))  
        else:  
            gal_Tb = self.galaxy_interp((ra, dec))
        gal_Tb = gal_Tb * (frequency / 1420)**-2.7 + self.Tcmb
        return gal_Tb

    @staticmethod
    def sun_brightness(frequency, year=2005): 
        """ Simple solar brightness temperature model using average solar flux
            value and sinusoidal dependence from Ho et al. 2008 JPL report. 

            :param frequency: Frequency in MHz 
            :param year: Observation year """

        wave = spc.c / (frequency * 1e6)
        F = 100 + 50 * np.cos(2 * np.pi / 11 * year)
        F = F * 1e-22  # Average solar flux in W/m2/Hz, 1e-22 is 'solar flux unit'
        ster_sun = 5.981149e-5  # Steradians 

        Tb = wave**2 * F / (2 * spc.k * ster_sun)

        return Tb

    @staticmethod
    def moon_brightness():
        """ Returns 275 K for moon brightness temperature
        """
        return 275  



from itertools import repeat
from multiprocessing import Pool
import copy
import numpy as np 
import pandas as pd 
from scipy.linalg import svd
from scipy.optimize import least_squares
import scipy.constants as spc
from cftime import date2num
import foam.geomag as geomag
from tqdm.auto import tqdm, trange


class solver():
    """ The solver object combines individual FOAM modules to form a full forward or inverse model. 
        The default for all inputs is 'None', and inputs that are not included in the solver 
        are not called. Appropriate defaults are used in their place.  
    
        Function versions are included to facilitate ease of use with the spacecraft object

        Module inputs: 
        :param ocean: an ocean object
        :param atmosphere: an atmosphere object 
        :param ionosphere: an ionosphere object 
        :param sky: a sky object
    """

    def __init__(self, ocean=None, atmosphere=None, ionosphere=None, sky=None, verbose=False): 

        self.ocean = ocean 
        self.atmosphere = atmosphere 
        self.ionosphere = ionosphere
        self.sky = sky 
        self.verbose = verbose

    def compute_TB(self, frequency, time, lat, lon, theta, phi, ra=None, dec=None, 
                   sun_flag=None, moon_flag=None, in_epoch=False, use_time=True):
        """ This function computes forward model brightness temperatures for the states of provided solver inputs

            Inputs: 
            :param frequency: Observing frequencies in MHz
            :param time: Time in reference coords
            :param lat: Latitudes in degrees 
            :param lon: Longitudes in degrees 
            :param theta: Surface elevation angle (zero is nadir) in degrees
            :param phi: Surface azimuth angle (zero is in orbit plane) in degrees 
            :param ra: Right ascension of reflected surface vector in degrees 
            :param dec: Declination of reflected surface vector in degrees
            :param sun_flag: Boolean array where sun is in path of reflected surface vector
            :param moon_flag: Boolean array where moon is in path of reflected surface vector

        """
        if not use_time: 
            time = self.ocean.datetime[0]
            in_epoch = False
        if not in_epoch: 
            time = date2num(pd.to_datetime(time).to_pydatetime(), self.ocean.time_reference)

        # Atmosphere 
        if self.atmosphere is not None: 
            if not use_time: 
                time = self.atmosphere.datetime[0]
                in_epoch = False 
            if not in_epoch: 
                time = date2num(pd.to_datetime(time).to_pydatetime(), self.atmosphere.time_reference)
            uwind = self.atmosphere.uwind_interp((time, lat, lon))
            vwind = self.atmosphere.vwind_interp((time, lat, lon))
            wind = np.sqrt(uwind**2 + vwind**2)
            tbup, tbdn, prop_dict = self.atmosphere.get_atmosphere_tb(frequency, time, lat, lon, theta, 
                                                                      in_epoch=in_epoch, use_time=use_time)
            transup = prop_dict['upward_transmissivity']
            year = np.mean(self.atmosphere.datetime.year.values)  # Need to adjust for time accuracy later
        else: 
            uwind = vwind = wind = np.zeros(np.shape(lat))
            tbup = tbdn = 0 
            transup = 1
            year = 2005

        # Ocean 
        if self.ocean is not None: 
            if not use_time: 
                time = self.ocean.datetime[0]
                in_epoch = False 
            if not in_epoch: 
                time = date2num(pd.to_datetime(time).to_pydatetime(), self.ocean.time_reference)
            o_tb, odict = self.ocean.get_ocean_TB(frequency, time, lat, lon, uwind, vwind, theta, phi, 
                                                  in_epoch=in_epoch, use_time=use_time)
            emis = odict['emissivity']
        else: 
            o_tb = np.zeros((4, np.shape(frequency), np.shape(lat)))
            emis = np.ones((4, np.shape(frequency), np.shape(lat)))
            emis[2:, :] = 0 
        
        # Sky 
        if self.sky is not None: 
            gal_tb = self.sky.galaxy_brightness(frequency, ra, dec, wind)
            gal_tb[:, moon_flag] = self.sky.moon_brightness() 
            gal_tb[:, sun_flag] = self.sky.sun_brightness(frequency, year)
        else: 
            gal_tb = 0

        TB = o_tb
        TB[:2] = o_tb[:2] * transup + tbup + (1 - emis[:2]) * (tbdn + gal_tb * transup) * transup
        
        # Ionosphere 
        if self.ionosphere is not None: 
            if not use_time: 
                time = self.ionosphere.datetime[0]
                in_epoch = False 
            if not in_epoch: 
                time = date2num(pd.to_datetime(time).to_pydatetime(), self.ionosphere.time_reference)
            far_angle = self.ionosphere.compute_faraday_angle(frequency, time, lat, lon, theta, phi, 
                                                              in_epoch=in_epoch, use_time=use_time)
            TB = self.ionosphere.faraday_rotation(-far_angle, TB)
    
        return np.real(TB)

    def compute_spacecraft_TB(self, frequency, obs_dict, use_time=True): 
        """ Initializes 'compute_TB' using the output from the spacecraft.make_obs_grid 
            function. 

            Inputs: 
            frequency - Observing frequencies in MHz 
            obs_dict - output of spacecraft.make_obs_grid
        """ 
        epoch = obs_dict['epoch']
        lat = obs_dict['lat']
        lon = obs_dict['lon']
        theta = obs_dict['theta']
        phi = obs_dict['phi']
        ra = obs_dict['ra']
        dec = obs_dict['dec']
        sun_flag = obs_dict['sun_flag']
        moon_flag = obs_dict['moon_flag']
        TB = self.compute_TB(frequency, epoch, lat, lon, theta, phi, ra, dec, sun_flag, moon_flag, in_epoch=True, use_time=use_time)
        anc_pack = self.ancillary_pack(epoch, lat, lon, theta, phi, ra=ra, dec=dec, in_epoch=True, use_time=use_time)
        return TB, anc_pack

    def compute_TB_map(self, frequency, theta=0, phi=0, mesh_spacing=1):
        """ Initializes compute_TB for a full map of the Earth's surface with arbitrary 
            pointing information. Sky brightness assumes nadir reflection, and the sun and moon are 
            not included. The map mesh spacing can be manually defined

            Inputs:
            :param frequency: Frequency in MHz (dimension O)
            :param theta: Elevation angle of spacecraft with respect to surface normal
            :param phi: Azimuth angle of spacecraft with respect to surface normal
            :param mesh_spacing: Spacing of latitude and longitude mesh grids

            :return: Polarimetric brightness temperatures 
                    (dimension OxMxN) (M and N correspond to latitude and longitude)
        """

        lat_flat = np.linspace(89.5, -89.5, int(180 / mesh_spacing))
        lon_flat = np.linspace(-179.5, 179.5, int(360 / mesh_spacing))
        lon, lat = np.meshgrid(lon_flat, lat_flat)
        thetas = theta * np.ones(np.shape(lon))
        phis = phi * np.ones(np.shape(lon))

        return self.compute_TB(frequency, lat, lon, thetas, phis, lon, lat)

    def ancillary_pack(self, time, lat, lon, theta, phi, ra=None, dec=None, use_time=True, in_epoch=False): 
        """ Assembles ancillary data from various objects in a form that can be read by the retrieval method 
        """ 

        anc_dict = {}
        anc_dict['lat'] = lat
        anc_dict['lon'] = lon 
        anc_dict['theta'] = theta
        anc_dict['phi'] = phi 
        if (ra is not None) and (dec is not None): 
            anc_dict['ra'] = ra 
            anc_dict['dec'] = dec

        if self.ocean is not None:
            if not use_time: 
                time = self.ocean.datetime[0]
                in_epoch = False 
            if not in_epoch: 
                time = date2num(pd.to_datetime(time).to_pydatetime(), self.ocean.time_reference)
            sst = self.ocean.sst_interp((time, lat, lon))
            sst[sst < 270] = np.nan  # Eliminates fill values
            anc_dict['sst'] = sst 
            sss = self.ocean.sss_interp((time, lat, lon))
            sss[sss <= 0] = np.nan  # Eliminate fill values
            anc_dict['sss'] = sss 
            landmask = self.ocean.landmask_interp((lat, lon))
            anc_dict['landmask'] = landmask
        else: 
            anc_dict['sst'] = np.zeros(np.shape(lat))
            anc_dict['sss'] = np.zeros(np.shape(lat))
            anc_dict['landmask'] = np.zeros(np.shape(lat))

        if self.atmosphere is not None: 
            if not use_time: 
                time = self.atmosphere.datetime[0]
                in_epoch = False 
            if not in_epoch: 
                time = date2num(pd.to_datetime(time).to_pydatetime(), self.atmosphere.time_reference)
            uwind = self.atmosphere.uwind_interp((time, lat, lon))
            anc_dict['uwind'] = uwind
            vwind = self.atmosphere.vwind_interp((time, lat, lon))
            anc_dict['vwind'] = vwind
            wind = np.sqrt(uwind**2 + vwind**2)
            anc_dict['windspd'] = wind
            winddir = np.arctan2(vwind, uwind)
            anc_dict['winddir'] = winddir
            prwtr = self.atmosphere.prwtr_interp((time, lat, lon))
            anc_dict['prwtr'] = prwtr
            if self.atmosphere.lwtr_interp is not None: 
                lwtr = self.atmosphere.lwtr_interp((time, lat, lon))
            else: 
                lwtr = np.zeros(np.shape(prwtr))
            anc_dict['lwtr'] = lwtr
            airtemp = self.atmosphere.airtemp_interp((time, lat, lon))
            anc_dict['airtemp'] = airtemp
            airpres = self.atmosphere.airpres_interp((time, lat, lon))
            anc_dict['srfpres'] = airpres
        else: 
            anc_dict['uwind'] = np.zeros(np.shape(lat))
            anc_dict['vwind'] = np.zeros(np.shape(lat))
            anc_dict['windspd'] = np.zeros(np.shape(lat))
            anc_dict['winddir'] = np.zeros(np.shape(lat))
            anc_dict['prwtr'] = np.zeros(np.shape(lat))
            anc_dict['lwtr'] = np.zeros(np.shape(lat))
            anc_dict['airtemp'] = np.zeros(np.shape(lat))
            anc_dict['srfpres'] = np.zeros(np.shape(lat))

        if self.ionosphere is not None: 
            if not use_time: 
                time = self.ionosphere.datetime[0]
                in_epoch = False 
            if not in_epoch: 
                time = date2num(pd.to_datetime(time).to_pydatetime(), self.ionosphere.time_reference)
            TEC = self.ionosphere.TEC_interp((time, lat, lon))
            anc_dict['tec'] = TEC
            gm = geomag.GeoMag(self.ionosphere.mag_file)
            mag = gm.GeoMag(lat, lon, h=(675e3 / spc.foot))
            x = np.sin(np.radians(theta)) * np.cos(np.radians(phi))
            y = np.sin(np.radians(theta)) * np.sin(np.radians(phi))
            z = np.cos(np.radians(theta)) 
            mag_field = -(mag.bx * x + mag.by * y + mag.bz * z)
            anc_dict['mag_field'] = mag_field 
        else: 
            anc_dict['tec'] = np.zeros(np.shape(lat))
            anc_dict['mag_field'] = np.zeros(np.shape(lat))

        if self.sky is not None: 
            wind = anc_dict['windspd']
            coldsky = self.sky.galaxy_interp((wind, ra, dec))
            anc_dict['coldsky'] = coldsky
        else: 
            anc_dict['coldsky'] = np.zeros(np.shape(lat))

        return anc_dict

    def compute_manual_TB(self, frequency=1e3, lat=0, theta=0, phi=0, 
                          sst=300, sss=35, uwind=5, vwind=5, prwtr=0, lwtr=0, airtemp=0, srfpres=0, 
                          tec=None, coldsky=0, mag_field=0):
        """ This routine can be used for quick calculations of brightness temperature over the ocean, 
            and is used by the inversion routines to retrieve state variables from measurements

            Inputs: 
            :param frequency: Frequency in MHz
            :param sst: Sea surface temperature in Kelvin
            :param sss: Sea surface salinity in psu
            :param uwind: 10 meter U wind in m/s
            :param vwind: 10 meter V wind in m/s
            :param tau: Atmospheric optical depth (only if retrieved)
            :param tec: Ionospheric TEC (only if retrieved)
            :param coldsky: Galactic emission temperature in K
            :param lat: Latitude in degrees north
            :param theta: Incidence angle in degrees 
            :param phi: Azimuth angle in degrees 
            :param prwtr: Precipitable water vapor in kg/kg (for quick tau calculations)
            :param mag_field: Scalar magnetic field in nT

            """
        # print('Arg Stack')
        # print('frequency:{}'.format(frequency))
        # print('lat: {}'.format(lat))
        # print('theta: {}'.format(theta))
        # print('phi: {}'.format(phi))
        # print('sst: {}'.format(sst))
        # print('sss: {}'.format(sss))
        # print('uwind: {}'.format(uwind))
        # print('vwind: {}'.format(vwind))
        # print('prwtr: {}'.format(prwtr))
        # print('lwtr: {}'.format(lwtr))
        # print('airtemp: {}'.format(airtemp))
        # print('srfpres: {}'.format(srfpres))
        # print('tec: {}'.format(tec))
        # print('coldsky: {}'.format(coldsky))
        # print('mag_field: {}'.format(mag_field))

        ghz = frequency / 1e3

        # Ocean surface
        emis = self.ocean.get_ocean_emissivity(frequency, sst, sss, uwind, vwind, theta, phi)
        
        # Propagation
        # Old propagation parameterization 
        # tau_coeff = np.array([0.013664614494878, -1.83402804591467e-5, -1.02330938120674e-5])
        # tau = tau_coeff[0] + tau_coeff[1] * sst + tau_coeff[2] * prwtr
        # Teff_coeff = np.array([-7.69054064356577e-10, 4.05537390612154e-7, 2.68693677239213e-6, 
        #                     -0.00770149860447519, -0.0576638279677582, 272.404431351231])
        # Teff = np.polyval(Teff_coeff, lat)

        if self.atmosphere is not None: 
            angle = np.radians(theta)
            tbup, tbdn, prop_dict = self.atmosphere.get_atmosphere_prop(frequency, prwtr, lwtr, airtemp, srfpres, lat, angle)
            tau = prop_dict['optical_depth']
            Teff = prop_dict['effective_temperature']
            trans = np.exp(-tau / np.cos(angle))
        else: 
            trans = 1
            Teff = 0 

        if self.sky is not None: 
            Tsky = coldsky * (ghz / 1.42)**-2.7 + self.sky.Tcmb
        else: 
            Tsky = 0

        Tup = (1 - trans) * Teff
        Tdown = Tsky * trans + (1 - trans) * Teff

        TB = emis * sst 
        # Fix proper polarimetric behaviour later 
        TB[:2] = emis[:2] * sst * trans + Tup + (1 - emis[:2]) * Tdown * trans
        if self.ionosphere is not None: 
            far_angle = np.radians(1.3549e-5 / ghz**2 * tec * mag_field)
            TB = self.ionosphere.faraday_rotation(-far_angle, TB)

        return np.real(TB)

    def retrieval(self, TB, anc_dict,  # Positional information 
                  frequency=np.array([1.4e3]), bandwidth=np.array([10]), 
                  noise_figure=1, int_time=1, cal_noise=0,  # Instrument parameters 
                  retrieve=['sss', 'sst', 'windspd', 'prwtr', 'lwtr', 'tec', 'coldsky'],  # Retrieval arguments
                  options=None):  # Subselect TB data range  
        """ Inverse model to determine geophysical parameters from brightness temperature data. 
            The user can customize the retrieval process in several ways, as explained by the parameters 

            :param TB: Brightness temperature array with dimension MxNxO
                       M - Dimension of Stokes vector (TBV, TBH, U, V)
                       N - Number of frequency channels 
                       O - Number of latitude/longitude pairs
            :param bandwidth: Bandwidth in MHz 
            :param noise_figure: Receiver noise figure in dB
            :integration_time: Integration time in seconds 
            :param cal_noise: Optional noise source that can be added to NEDT 
            :param retrieve: A list of strings containing parameters that should be retrieved
                             Options are: 'sst', 'sss', 'windspd', 'tau', 'tec' and 'coldsky'

            Note: An alternate versioning scheme for TB is
            TB: Brightness temperature array with dimension MxNxOxP
                       M - Dimension of Stokes vector (TBV, TBH, U, V)
                       N - Number of frequency channels 
                       O - Number of radiometer horns 
                       P - Number of latitude/longitude pairs
        """ 

        if options is None: 
            options = {} 
        options.setdefault('ancillary_guess', True)

        # Flatten horns
        # anc_dict = {k: v.flatten() for k, v in anc_dict.items()}
        # shape = np.shape(TB)
        # TB = TB.reshape(*shape[:-2], shape[-2] * shape[-1])

        out_dict, unc_dict, anc_dict = self._retrieve(self, TB, anc_dict, frequency, bandwidth, 
                                                      noise_figure, int_time, cal_noise, 
                                                      retrieve, options)
        
        return out_dict, unc_dict, anc_dict

    def parallel_retrieval(self, TB, anc_dict,  # Positional information 
                           frequency=np.array([1.4e3]), bandwidth=np.array([10]), 
                           noise_figure=1, int_time=1, cal_noise=0,  # Instrument parameters 
                           retrieve=['sss', 'sst', 'windspd', 'prwtr', 'lwtr', 'tec', 'coldsky'],  # Retrieval arguments
                           options=None, ndiv=4, pool=None, nproc=4):  # Subselect TB data range  
        """ Inverse model to determine geophysical parameters from brightness temperature data. 
            Uses multiprocessing

            :param TB: Brightness temperature array with dimension MxNxO
                       M - Dimension of Stokes vector (TBV, TBH, U, V)
                       N - Number of frequency channels 
                       O - Number of latitude/longitude pairs
            :param bandwidth: Bandwidth in MHz 
            :param noise_figure: Receiver noise figure in dB
            :integration_time: Integration time in seconds 
            :param cal_noise: Optional noise source that can be added to NEDT 
            :param retrieve: A list of strings containing parameters that should be retrieved
                             Options are: 'sst', 'sss', 'windspd', 'tau', 'tec' and 'coldsky'
        """ 

        if options is None: 
            options = {} 
        options.setdefault('ancillary_guess', True)

        # Flatten horns
        # anc_dict = {k: v.flatten() for k, v in anc_dict.items()}
        # shape = np.shape(TB)
        # TB = TB.reshape(*shape[:-2], shape[-2] * shape[-1])
        
        # Partition inputs
        TB_list = np.array_split(TB, ndiv, axis=-1)
        anc_dict_list = [{} for x in range(ndiv)]
        for k, v in anc_dict.items():
            val_list = np.array_split(v, ndiv, axis=-1)
            for i, ad in enumerate(anc_dict_list): 
                ad[k] = val_list[i]

        if not pool: 
            local_pool = True 
            pool = Pool(processes=nproc)
        else: 
            local_pool = False

        # Build solver copy queue
        solver_list = [] 
        for n in range(nproc): 
            solver_list.append(copy.deepcopy(self))
        solver_list.extend(solver_list * (ndiv - 1))

        zipper = list(zip(solver_list, TB_list, anc_dict_list, 
                          repeat(frequency), repeat(bandwidth), 
                          repeat(noise_figure), repeat(int_time),
                          repeat(cal_noise), repeat(retrieve), repeat(options), range(ndiv)))

        tuple_dicts = pool.starmap(self._retrieve, zipper)

        if local_pool: 
            pool.close()
            pool.join()
        
        tuple_dicts = tuple(zip(*tuple_dicts))  # Reorganize by element
        out_dict_list = []
        for td in tuple_dicts: 
            keys = td[0].keys()
            merged_dict = {k: np.concatenate([d.get(k) for d in td]) for k in keys}
            out_dict_list.append(merged_dict)
        out_dict, unc_dict, anc_dict = out_dict_list

        return out_dict, unc_dict, anc_dict

    @staticmethod
    def _retrieve(solver, TB, anc_dict,  # Positional information 
                  frequency, bandwidth, noise_figure, int_time, cal_noise,  # Instrument parameters 
                  retrieve, options, ith=0): 

        # Generate error estimates 
        Trec = 290 * (10**(noise_figure / 10) - 1) 
        Tsys = TB + Trec
        bandwidth = bandwidth[np.newaxis, :, np.newaxis]
        NEDT = Tsys / np.sqrt(bandwidth * 1e6 * int_time)
        tot_error = np.real(np.sqrt(NEDT**2 + cal_noise**2))
        # print('NEDT: {}'.format(np.nanmean(NEDT).flatten()))

        # Setup ancillary data inputs 
        
        low_bounds = []
        high_bounds = [] 
        guess = {}
        good_guess = {}
        if 'sss' in retrieve: 
            low_bounds.append(0)
            high_bounds.append(50)
            guess['sss'] = 34
            good_guess['sss'] = anc_dict['sss']
        if 'sst' in retrieve: 
            low_bounds.append(260)
            high_bounds.append(320)
            guess['sst'] = 293.15
            good_guess['sst'] = anc_dict['sst']
        if 'windspd' in retrieve: 
            low_bounds.append(0)
            high_bounds.append(20)
            guess['windspd'] = 5
            good_guess['windspd'] = anc_dict['windspd']
        if 'prwtr' in retrieve: 
            low_bounds.append(0)
            high_bounds.append(100)
            guess['prwtr'] = 1e-2
            good_guess['prwtr'] = anc_dict['prwtr']
        if 'lwtr' in retrieve: 
            low_bounds.append(0)
            high_bounds.append(100)
            guess['lwtr'] = 1e-2
            good_guess['lwtr'] = anc_dict['lwtr']
        if 'tec' in retrieve: 
            low_bounds.append(0)
            high_bounds.append(100)
            guess['tec'] = 1
            good_guess['tec'] = anc_dict['tec']
        if 'coldsky' in retrieve: 
            low_bounds.append(0)
            high_bounds.append(20)
            guess['coldsky'] = 2
            good_guess['coldsky'] = anc_dict['coldsky']

        bnds = (np.array(low_bounds), np.array(high_bounds))
        guess_array = np.array(list(guess.values()))
        labels = list(guess.keys())
        lat = anc_dict['lat']
        out = np.zeros((len(guess), len(lat)))
        uncertain = np.zeros((len(guess), len(lat)))
        for i in trange(len(lat), colour='green', desc='Thread {}'.format(ith), 
                        mininterval=0.1, maxinterval=1., miniters=1): 
            if not np.isnan(TB[..., i]).any(): 
                y_err = TB[..., i] + np.random.randn(*np.shape(TB[..., i])) * tot_error[..., i]
                argpack = [frequency, y_err, tot_error[..., i], anc_dict, i]

                if options['ancillary_guess']: 
                    start = np.array([good_guess[x][i] for x in good_guess.keys()])
                else: 
                    start = guess_array
                args = (labels, argpack)
                if solver.verbose: 
                    print('Starting Retrieval  #############')
                    print('Guess: {}'.format(start))

                start_scale = start 
                start_scale[start_scale == 0] = 1

                try:
                    res = least_squares(solver._cf, start, method='trf', x_scale=start_scale, args=args,
                                    bounds=bnds, max_nfev=1e10, ftol=1e-4, xtol=1e-4)
                    out[:, i] = res.x

                    # Compute std. dev. using scipy curve_fit algorithm 
                    _, s, VT = svd(res.jac, full_matrices=False)
                    threshold = np.finfo(float).eps * max(res.jac.shape) * s[0]
                    s = s[s > threshold]
                    VT = VT[:s.size]
                    pcov = np.dot(VT.T / s**2, VT)
                    uncertain[:, i] = np.sqrt(np.diag(pcov))

                except ValueError: 
                    out[:, i] = np.nan
                    uncertain[:, i] = np.nan

                if solver.verbose: 
                    print('Ending Retrieval #############')
                    print('Result: {}'.format(res.x))

            else: 
                out[:, i] = np.nan
        out_dict = dict(zip(guess.keys(), out))
        unc_dict = dict(zip(guess.keys(), uncertain))
        return out_dict, unc_dict, anc_dict

    def _cf(self, guess, labels, argpack): 

        # Unpack argpack 
        frequency = argpack[0]
        y_err = argpack[1]
        tot_error = argpack[2]
        anc_dict = argpack[3]
        index = argpack[4]

        if 'sss' in labels: 
            sss = guess[labels.index('sss')]
        else: 
            sss = anc_dict['sss'][index] 
        if 'sst' in labels: 
            sst = guess[labels.index('sst')]
        else: 
            sst = anc_dict['sst'][index]
        if 'windspd' in labels: 
            windspd = guess[labels.index('windspd')]
        else: 
            windspd = anc_dict['windspd'][index]
        if 'prwtr' in labels: 
            prwtr = guess[labels.index('prwtr')]
        else: 
            prwtr = anc_dict['prwtr'][index]
        if 'lwtr' in labels: 
            lwtr = guess[labels.index('lwtr')]
        else: 
            lwtr = anc_dict['lwtr'][index]            
        if 'tec' in labels: 
            tec = guess[labels.index('tec')]
        else: 
            tec = anc_dict['tec'][index]
        if 'coldsky' in labels: 
            coldsky = guess[labels.index('coldsky')]
        else: 
            coldsky = anc_dict['coldsky'][index]

        winddir = anc_dict['winddir'][index]
        uwind = windspd * np.cos(winddir)
        vwind = windspd * np.sin(winddir)        
        ret_y = self.compute_manual_TB(frequency=frequency, lat=anc_dict['lat'][index], 
                                       theta=anc_dict['theta'][index], phi=anc_dict['phi'][index], 
                                       sst=sst, sss=sss, uwind=uwind, vwind=vwind, prwtr=prwtr, lwtr=lwtr, 
                                       airtemp=anc_dict['airtemp'][index], srfpres=anc_dict['srfpres'][index], 
                                       tec=tec, coldsky=coldsky, mag_field=anc_dict['mag_field'][index])

        cost = (ret_y.ravel() - y_err.ravel()) / tot_error.ravel()
        return cost


def bin_observations(vals, obs_dict, lat_bins, lon_bins, epoch_bins=None): 
    pd_dict = obs_dict.copy()
    pd_dict['vals'] = vals
    nset = pd.DataFrame.from_dict(pd_dict).dropna()
    if epoch_bins is not None: 
        nset = nset[['lat', 'lon', 'epoch', 'vals']]    
    else: 
        nset = nset[['lat', 'lon', 'vals']]

    # Binning 
    lat_dex = np.digitize(nset['lat'], lat_bins)
    lon_dex = np.digitize(nset['lon'], lon_bins)
    nset['lat'] = lat_bins[lat_dex - 1]
    nset['lon'] = lon_bins[lon_dex - 1]
    if epoch_bins is not None: 
        ep_dex = np.digitize(nset['epoch'], epoch_bins)
        nset['epoch'] = epoch_bins[ep_dex - 1]
        nsgb = nset.groupby(['lat', 'lon', 'epoch'])
    else:    
        nsgb = nset.groupby(['lat', 'lon'])
    
    mean = nsgb.mean()
    std = nsgb.std()

    # Conform bins
    if epoch_bins is not None: 
        rdex = pd.MultiIndex.from_product([lat_bins, lon_bins, epoch_bins], names=['lat', 'lon', 'epoch'])
    else: 
        rdex = pd.MultiIndex.from_product([lat_bins, lon_bins], names=['lat', 'lon'])
    mean = mean.reindex(rdex)
    std = std.reindex(rdex)

    return mean, std


